import time

import numpy as np

import torch

from ..attack import Attack


class APGDT(Attack):
    r"""
    APGD-Targeted in the paper 'Reliable evaluation of adversarial robustness with an ensemble of diverse parameter-free attacks.'
    Targeted attack for every wrong classes.
    [https://arxiv.org/abs/2003.01690]
    [https://github.com/fra31/auto-attack]

    Distance Measure : Linf, L2

    Arguments:
        model (nn.Module): model to attack.
        norm (str): Lp-norm of the attack. ['Linf', 'L2'] (Default: 'Linf')
        eps (float): maximum perturbation. (Default: 8/255)
        steps (int): number of steps. (Default: 10)
        n_restarts (int): number of random restarts. (Default: 1)
        seed (int): random seed for the starting point. (Default: 0)
        eot_iter (int): number of iteration for EOT. (Default: 1)
        rho (float): parameter for step-size update (Default: 0.75)
        verbose (bool): print progress. (Default: False)
        n_classes (int): number of classes. (Default: 10)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.APGDT(model, norm='Linf', eps=8/255, steps=10, n_restarts=1, seed=0, eot_iter=1, rho=.75, verbose=False, n_classes=10)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        norm="Linf",
        eps=8 / 255,
        steps=10,
        n_restarts=1,
        seed=0,
        eot_iter=1,
        rho=0.75,
        verbose=False,
        n_classes=10,
    ):
        super().__init__("APGDT", model)
        self.eps = eps
        self.steps = steps
        self.norm = norm
        self.n_restarts = n_restarts
        self.seed = seed
        self.eot_iter = eot_iter
        self.thr_decr = rho
        self.verbose = verbose
        self.target_class = None
        self.n_target_classes = n_classes - 1
        self.supported_mode = ["default"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)
        _, adv_images = self.perturb(images, labels, cheap=True)

        return adv_images

    def check_oscillation(self, x, j, k, y5, k3=0.5):
        t = np.zeros(x.shape[1])
        for counter5 in range(k):
            t += x[j - counter5] > x[j - counter5 - 1]

        return t <= k * k3 * np.ones(t.shape)

    def check_shape(self, x):
        return x if len(x.shape) > 0 else np.expand_dims(x, 0)

    def dlr_loss_targeted(self, x, y, y_target):
        x_sorted, ind_sorted = x.sort(dim=1)

        return -(x[np.arange(x.shape[0]), y] - x[np.arange(x.shape[0]), y_target]) / (
            x_sorted[:, -1] - 0.5 * x_sorted[:, -3] - 0.5 * x_sorted[:, -4] + 1e-12
        )

    def attack_single_run(self, x_in, y_in):
        x = x_in.clone() if len(x_in.shape) == 4 else x_in.clone().unsqueeze(0)
        y = y_in.clone() if len(y_in.shape) == 1 else y_in.clone().unsqueeze(0)

        self.steps_2, self.steps_min, self.size_decr = (
            max(int(0.22 * self.steps), 1),
            max(int(0.06 * self.steps), 1),
            max(int(0.03 * self.steps), 1),
        )  # nopep8
        if self.verbose:
            print(
                "parameters: ", self.steps, self.steps_2, self.steps_min, self.size_decr
            )  # nopep8

        if self.norm == "Linf":
            t = 2 * torch.rand(x.shape).to(self.device).detach() - 1
            x_adv = x.detach() + self.eps * torch.ones([x.shape[0], 1, 1, 1]).to(
                self.device
            ).detach() * t / (
                t.reshape([t.shape[0], -1])
                .abs()
                .max(dim=1, keepdim=True)[0]
                .reshape([-1, 1, 1, 1])
            )  # nopep8
        elif self.norm == "L2":
            t = torch.randn(x.shape).to(self.device).detach()
            x_adv = x.detach() + self.eps * torch.ones([x.shape[0], 1, 1, 1]).to(
                self.device
            ).detach() * t / (
                (t ** 2).sum(dim=(1, 2, 3), keepdim=True).sqrt() + 1e-12
            )  # nopep8
        x_adv = x_adv.clamp(0.0, 1.0)
        x_best = x_adv.clone()
        x_best_adv = x_adv.clone()
        loss_steps = torch.zeros([self.steps, x.shape[0]])
        loss_best_steps = torch.zeros([self.steps + 1, x.shape[0]])
        acc_steps = torch.zeros_like(loss_best_steps)

        output = self.get_logits(x)
        y_target = output.sort(dim=1)[1][:, -self.target_class]

        x_adv.requires_grad_()
        grad = torch.zeros_like(x)
        for _ in range(self.eot_iter):
            with torch.enable_grad():
                # 1 forward pass (eot_iter = 1)
                logits = self.get_logits(x_adv)
                loss_indiv = self.dlr_loss_targeted(logits, y, y_target)
                loss = loss_indiv.sum()

            # 1 backward pass (eot_iter = 1)
            grad += torch.autograd.grad(loss, [x_adv])[0].detach()

        grad /= float(self.eot_iter)
        grad_best = grad.clone()

        acc = logits.detach().max(1)[1] == y
        acc_steps[0] = acc + 0
        loss_best = loss_indiv.detach().clone()

        step_size = (
            self.eps
            * torch.ones([x.shape[0], 1, 1, 1]).to(self.device).detach()
            * torch.Tensor([2.0]).to(self.device).detach().reshape([1, 1, 1, 1])
        )  # nopep8
        x_adv_old = x_adv.clone()
        # counter = 0
        k = self.steps_2 + 0
        u = np.arange(x.shape[0])
        counter3 = 0

        loss_best_last_check = loss_best.clone()
        reduced_last_check = np.zeros(loss_best.shape) == np.zeros(loss_best.shape)

        # n_reduced = 0
        for i in range(self.steps):
            # gradient step
            with torch.no_grad():
                x_adv = x_adv.detach()
                grad2 = x_adv - x_adv_old
                x_adv_old = x_adv.clone()

                a = 0.75 if i > 0 else 1.0

                if self.norm == "Linf":
                    x_adv_1 = x_adv + step_size * torch.sign(grad)
                    x_adv_1 = torch.clamp(
                        torch.min(torch.max(x_adv_1, x - self.eps), x + self.eps),
                        0.0,
                        1.0,
                    )  # nopep8
                    x_adv_1 = torch.clamp(
                        torch.min(
                            torch.max(
                                x_adv + (x_adv_1 - x_adv) * a + grad2 * (1 - a),
                                x - self.eps,
                            ),
                            x + self.eps,
                        ),
                        0.0,
                        1.0,
                    )  # nopep8

                elif self.norm == "L2":
                    x_adv_1 = x_adv + step_size[0] * grad / (
                        (grad ** 2).sum(dim=(1, 2, 3), keepdim=True).sqrt() + 1e-12
                    )  # nopep8
                    x_adv_1 = torch.clamp(
                        x
                        + (x_adv_1 - x)
                        / (
                            ((x_adv_1 - x) ** 2).sum(dim=(1, 2, 3), keepdim=True).sqrt()
                            + 1e-12
                        )
                        * torch.min(
                            self.eps * torch.ones(x.shape).to(self.device).detach(),
                            ((x_adv_1 - x) ** 2)
                            .sum(dim=(1, 2, 3), keepdim=True)
                            .sqrt(),
                        ),
                        0.0,
                        1.0,
                    )  # nopep8
                    x_adv_1 = x_adv + (x_adv_1 - x_adv) * a + grad2 * (1 - a)
                    x_adv_1 = torch.clamp(
                        x
                        + (x_adv_1 - x)
                        / (
                            ((x_adv_1 - x) ** 2).sum(dim=(1, 2, 3), keepdim=True).sqrt()
                            + 1e-12
                        )
                        * torch.min(
                            self.eps * torch.ones(x.shape).to(self.device).detach(),
                            ((x_adv_1 - x) ** 2).sum(dim=(1, 2, 3), keepdim=True).sqrt()
                            + 1e-12,
                        ),
                        0.0,
                        1.0,
                    )  # nopep8

                x_adv = x_adv_1 + 0.0

            # get gradient
            x_adv.requires_grad_()
            grad = torch.zeros_like(x)
            for _ in range(self.eot_iter):
                with torch.enable_grad():
                    # 1 forward pass (eot_iter = 1)
                    logits = self.get_logits(x_adv)
                    loss_indiv = self.dlr_loss_targeted(logits, y, y_target)
                    loss = loss_indiv.sum()

                # 1 backward pass (eot_iter = 1)
                grad += torch.autograd.grad(loss, [x_adv])[0].detach()

            grad /= float(self.eot_iter)

            pred = logits.detach().max(1)[1] == y
            acc = torch.min(acc, pred)
            acc_steps[i + 1] = acc + 0
            x_best_adv[(pred == 0).nonzero().squeeze()] = (
                x_adv[(pred == 0).nonzero().squeeze()] + 0.0
            )
            if self.verbose:
                print("iteration: {} - Best loss: {:.6f}".format(i, loss_best.sum()))

            # check step size
            with torch.no_grad():
                y1 = loss_indiv.detach().clone()
                loss_steps[i] = y1.cpu() + 0
                ind = (y1 > loss_best).nonzero().squeeze()
                x_best[ind] = x_adv[ind].clone()
                grad_best[ind] = grad[ind].clone()
                loss_best[ind] = y1[ind] + 0
                loss_best_steps[i + 1] = loss_best + 0

                counter3 += 1

                if counter3 == k:
                    fl_oscillation = self.check_oscillation(
                        loss_steps.detach().cpu().numpy(),
                        i,
                        k,
                        loss_best.detach().cpu().numpy(),
                        k3=self.thr_decr,
                    )  # nopep8
                    fl_reduce_no_impr = (~reduced_last_check) * (
                        loss_best_last_check.cpu().numpy() >= loss_best.cpu().numpy()
                    )  # nopep8
                    fl_oscillation = ~(~fl_oscillation * ~fl_reduce_no_impr)
                    reduced_last_check = np.copy(fl_oscillation)
                    loss_best_last_check = loss_best.clone()

                    if np.sum(fl_oscillation) > 0:
                        step_size[u[fl_oscillation]] /= 2.0
                        # n_reduced = fl_oscillation.astype(float).sum()

                        fl_oscillation = np.where(fl_oscillation)

                        x_adv[fl_oscillation] = x_best[fl_oscillation].clone()
                        grad[fl_oscillation] = grad_best[fl_oscillation].clone()

                    counter3 = 0
                    k = np.maximum(k - self.size_decr, self.steps_min)

        return x_best, acc, loss_best, x_best_adv

    def perturb(self, x_in, y_in, best_loss=False, cheap=True):
        assert self.norm in ["Linf", "L2"]
        x = x_in.clone() if len(x_in.shape) == 4 else x_in.clone().unsqueeze(0)
        y = y_in.clone() if len(y_in.shape) == 1 else y_in.clone().unsqueeze(0)

        adv = x.clone()
        acc = self.get_logits(x).max(1)[1] == y
        # loss = -1e10 * torch.ones_like(acc).float()
        if self.verbose:
            print(
                "-------------------------- running {}-attack with epsilon {:.4f} --------------------------".format(
                    self.norm, self.eps
                )
            )
            print("initial accuracy: {:.2%}".format(acc.float().mean()))
        startt = time.time()

        torch.random.manual_seed(self.seed)
        torch.cuda.random.manual_seed(self.seed)

        if not cheap:
            raise ValueError("not implemented yet")

        else:
            for target_class in range(2, self.n_target_classes + 2):
                self.target_class = target_class
                for counter in range(self.n_restarts):
                    ind_to_fool = acc.nonzero().squeeze()
                    if len(ind_to_fool.shape) == 0:
                        ind_to_fool = ind_to_fool.unsqueeze(0)
                    if ind_to_fool.numel() != 0:
                        x_to_fool, y_to_fool = (
                            x[ind_to_fool].clone(),
                            y[ind_to_fool].clone(),
                        )  # nopep8
                        (
                            best_curr,
                            acc_curr,
                            loss_curr,
                            adv_curr,
                        ) = self.attack_single_run(
                            x_to_fool, y_to_fool
                        )  # nopep8
                        ind_curr = (acc_curr == 0).nonzero().squeeze()

                        acc[ind_to_fool[ind_curr]] = 0
                        adv[ind_to_fool[ind_curr]] = adv_curr[ind_curr].clone()
                        if self.verbose:
                            print(
                                "restart {} - target_class {} - robust accuracy: {:.2%} at eps = {:.5f} - cum. time: {:.1f} s".format(
                                    counter,
                                    self.target_class,
                                    acc.float().mean(),
                                    self.eps,
                                    time.time() - startt,
                                )
                            )

        return acc, adv



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

from airtk.defense.adt import AdtTraining
from airtk.defense.adtpp import AdtppTraining
from airtk.defense.currat import CurratTraining
from airtk.defense.fat import (
    FatTraining,
)
from airtk.defense.feature_scatter import (
    FeatureScatterTraining,
)
from airtk.defense.gairat import (
    GairatTraining,
)
from airtk.defense.trades import (
    TradesTraining,
)
from airtk.defense.tradesawp import TradesawpTraining
from airtk.defense.va import VaTraining
from airtk.defense.yopo import YopoTraining

__all__ = [
    "AdtTraining",
    "AdtppTraining",
    "CurratTraining",
    "FatTraining",
    "FeatureScatterTraining",
    "GairatTraining",
    "TradesTraining",
    "TradesawpTraining",
    "VaTraining",
    "YopoTraining",
]



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

from airtk.model.resnet import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
)
from airtk.model.wideresnet import WideResNet

__all__ = [
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152",
    "WideResNet",
]



# Modified from https://github.com/bearpaw/pytorch-classification/blob/master/models/cifar/wrn.py
import torch
import torch.nn as nn


class Normalize(nn.Module):
    def __init__(self, mean, std):
        super().__init__()
        assert len(mean) == len(std)
        self.n_channels = len(mean)
        self.register_buffer("mean", torch.tensor(mean))
        self.register_buffer("std", torch.tensor(std))

    def __str__(self):
        return "Normalize(" + "mean={}, std={}".format(self.mean, self.std) + ")"

    def extra_repr(self):
        return "mean={}, std={}".format(self.mean, self.std)

    def forward(self, inputs):
        # Broadcasting
        mean = self.mean.reshape(1, self.n_channels, 1, 1).to(inputs.device)
        std = self.std.reshape(1, self.n_channels, 1, 1).to(inputs.device)
        return (inputs - mean) / std


class InverseNormalize(nn.Module):
    def __init__(self, mean, std):
        super().__init__()
        assert len(mean) == len(std)
        self.n_channels = len(mean)
        self.register_buffer("mean", torch.tensor(mean))
        self.register_buffer("std", torch.tensor(std))

    def __str__(self):
        return "InverseNormalize(" + "mean={}, std={}".format(self.mean, self.std) + ")"

    def extra_repr(self):
        return "mean={}, std={}".format(self.mean, self.std)

    def forward(self, inputs):
        # Broadcasting
        mean = self.mean.reshape(1, self.n_channels, 1, 1).to(inputs.device)
        std = self.std.reshape(1, self.n_channels, 1, 1).to(inputs.device)
        return (inputs * std) + mean



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

# from models import *
from torch.autograd import Variable


def cwloss(output, target, confidence=50, num_classes=10):
    # Compute the probability of the label class versus the maximum other
    # The same implementation as in repo CAT https://github.com/sunblaze-ucb/curriculum-adversarial-training-CAT
    target = target.data
    target_onehot = torch.zeros(target.size() + (num_classes,))
    target_onehot = target_onehot.cuda()
    target_onehot.scatter_(1, target.unsqueeze(1), 1.0)
    target_var = Variable(target_onehot, requires_grad=False)
    real = (target_var * output).sum(1)
    other = ((1.0 - target_var) * output - target_var * 10000.0).max(1)[0]
    loss = -torch.clamp(
        real - other + confidence, min=0.0
    )  # equiv to max(..., 0.)
    loss = torch.sum(loss)
    return loss


# Geometry-aware projected gradient descent (GA-PGD)
def GA_PGD(
    model,
    data,
    target,
    epsilon,
    step_size,
    num_steps,
    loss_fn,
    category,
    rand_init,
):
    model.eval()
    Kappa = torch.zeros(len(data))
    if category == "trades":
        x_adv = (
            data.detach() + 0.001 * torch.randn(data.shape).cuda().detach()
            if rand_init
            else data.detach()
        )
        nat_output = model(data)
    if category == "Madry":
        x_adv = (
            data.detach()
            + torch.from_numpy(
                np.random.uniform(-epsilon, epsilon, data.shape)
            )
            .float()
            .cuda()
            if rand_init
            else data.detach()
        )
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    for k in range(num_steps):
        x_adv.requires_grad_()
        output = model(x_adv)
        predict = output.max(1, keepdim=True)[1]
        # Update Kappa
        for p in range(len(x_adv)):
            if predict[p] == target[p]:
                Kappa[p] += 1
        model.zero_grad()
        with torch.enable_grad():
            if loss_fn == "cent":
                loss_adv = nn.CrossEntropyLoss(reduction="mean")(
                    output, target
                )
            if loss_fn == "cw":
                loss_adv = cwloss(output, target)
            if loss_fn == "kl":
                criterion_kl = nn.KLDivLoss(size_average=False).cuda()
                loss_adv = criterion_kl(
                    F.log_softmax(output, dim=1), F.softmax(nat_output, dim=1)
                )
        loss_adv.backward()
        eta = step_size * x_adv.grad.sign()
        # Update adversarial data
        x_adv = x_adv.detach() + eta
        x_adv = torch.min(torch.max(x_adv, data - epsilon), data + epsilon)
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    x_adv = Variable(x_adv, requires_grad=False)
    return x_adv, Kappa


def eval_clean(model, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.cuda(), target.cuda()
            output = model(data)
            test_loss += F.cross_entropy(
                output, target, size_average=False
            ).item()
            pred = output.max(1, keepdim=True)[1]
            correct += pred.eq(target.view_as(pred)).sum().item()
    test_loss /= len(test_loader.dataset)
    test_accuracy = correct / len(test_loader.dataset)
    return test_loss, test_accuracy


def eval_robust(
    model,
    test_loader,
    perturb_steps,
    epsilon,
    step_size,
    loss_fn,
    category,
    random,
):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.enable_grad():
        for data, target in test_loader:
            data, target = data.cuda(), target.cuda()
            x_adv, _ = GA_PGD(
                model,
                data,
                target,
                epsilon,
                step_size,
                perturb_steps,
                loss_fn,
                category,
                rand_init=random,
            )
            output = model(x_adv)
            test_loss += F.cross_entropy(
                output, target, size_average=False
            ).item()
            pred = output.max(1, keepdim=True)[1]
            correct += pred.eq(target.view_as(pred)).sum().item()
    test_loss /= len(test_loader.dataset)
    test_accuracy = correct / len(test_loader.dataset)
    return test_loss, test_accuracy



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable


def quantization(arr, k):  # 0.5 will be rounded to floor
    quant = np.zeros(arr.shape)
    for i in range(1, k):
        # idx = np.where(arr>1.0*i/k)
        idx = np.where(arr >= 1.0 * i / k)
        quant[idx] += 1
    return quant / (k - 1)


def cwloss(output, target, confidence=50, num_classes=10):
    # compute the probability of the label class versus the maximum other
    target = target.data
    target_onehot = torch.zeros(target.size() + (num_classes,))
    target_onehot = target_onehot.cuda()
    target_onehot.scatter_(1, target.unsqueeze(1), 1.0)
    target_var = Variable(target_onehot, requires_grad=False)
    real = (target_var * output).sum(1)
    other = ((1.0 - target_var) * output - target_var * 10000.0).max(1)[0]
    loss = -torch.clamp(
        real - other + confidence, min=0.0
    )  # equiv to max(..., 0.)
    loss = torch.sum(loss)
    return loss


class LinfPGDAttack:
    def __init__(
        self,
        model,
        epsilon,
        num_steps,
        step_size,
        random_start,
        loss_func,
        GPU=True,
    ):
        """Attack parameter initialization. The attack performs k steps of
        size a, while always staying within epsilon from the initial
        point."""
        self.model = model
        self.epsilon = epsilon
        self.num_steps = num_steps
        self.step_size = step_size
        self.rand = random_start
        self.GPU = GPU

        if loss_func == "xent":
            self.criterion = nn.CrossEntropyLoss()
        elif loss_func == "cw":
            self.criterion = cwloss
        else:
            print("Unknown loss function. Defaulting to cross-entropy")
            self.criterion = nn.CrossEntropyLoss()

    def perturb_true(self, x_nat, y, k):
        """Given a set of examples (x_nat, y), returns a set of adversarial
        examples within epsilon of x_nat in l_infinity norm."""
        step_size = (self.step_size * self.num_steps / k) * np.random.uniform(
            1, 10.0 / k
        )
        # step_size =  self.step_size * self.num_steps / k if np.random.random()>0.5 else 2*self.epsilon
        if self.GPU:
            x_nat = x_nat.data.cpu().numpy()
            # self.model.cuda()
        else:
            x_nat = x_nat.data.numpy()
        if self.rand:
            x = x_nat + np.random.uniform(
                -self.epsilon, self.epsilon, x_nat.shape
            )
        else:
            x = np.copy(x_nat)

        for i in range(k):
            if self.GPU:
                x = torch.Tensor(x)
                x = x.cuda()
            else:
                x = torch.Tensor(x)
            x = Variable(x, requires_grad=True)
            self.model.zero_grad()
            output = self.model(x)
            loss = self.criterion(output, y)
            loss.backward()
            if self.GPU:
                grad = x.grad.data.cpu().numpy()
                x = x.data.cpu().numpy()
            else:
                grad = x.grad.data.numpy()
                x = x.data.numpy()

            x = np.add(x, step_size * (np.sign(grad)), out=x, casting="unsafe")

            x = np.clip(x, x_nat - self.epsilon, x_nat + self.epsilon)
            x = np.clip(x, 0, 1)  # ensure valid pixel range
        return Variable(torch.Tensor(x).cuda(), volatile=False)

    def perturb_quad(self, x_nat, y, k, level):
        """Given a set of examples (x_nat, y), returns a set of adversarial
        examples within epsilon of x_nat in l_infinity norm."""
        step_size = self.step_size * self.num_steps / k
        if self.GPU:
            x_nat = x_nat.data.cpu().numpy()
        #            self.model.cuda()
        else:
            x_nat = x_nat.data.numpy()
        if self.rand:
            x = x_nat + np.random.uniform(
                -self.epsilon, self.epsilon, x_nat.shape
            )
        else:
            x = np.copy(x_nat)

        for i in range(k):
            if self.GPU:
                x = torch.Tensor(x)
                x = x.cuda()
            else:
                x = torch.Tensor(x)
            x = Variable(x, requires_grad=True)
            self.model.zero_grad()
            output = self.model(x)
            loss = self.criterion(output, y)
            loss.backward()
            if self.GPU:
                grad = x.grad.data.cpu().numpy()
                x = x.data.cpu().numpy()
            else:
                grad = x.grad.data.numpy()
                x = x.data.numpy()

            x = np.add(x, step_size * (np.sign(grad)), out=x, casting="unsafe")

            x = np.clip(x, x_nat - self.epsilon, x_nat + self.epsilon)
            x = np.clip(x, 0, 1)  # ensure valid pixel range
        x_qt = quantization(x, level)
        return Variable(torch.Tensor(x).cuda(), volatile=False), Variable(
            torch.Tensor(x_qt).cuda(), volatile=False
        )



import time
from collections import OrderedDict

import torch
from torch.utils.data import DataLoader, TensorDataset


def wrapper_method(func):
    def wrapper_func(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        for atk in self.__dict__.get("_attacks").values():
            eval("atk." + func.__name__ + "(*args, **kwargs)")
        return result

    return wrapper_func


class Attack(object):
    r"""
    Base class for all attacks.

    .. note::
        It automatically set device to the device where given model is.
        It basically changes training mode to eval during attack process.
        To change this, please see `set_model_training_mode`.
    """

    def __init__(self, name, model):
        r"""
        Initializes internal attack state.

        Arguments:
            name (str): name of attack.
            model (torch.nn.Module): model to attack.
        """

        self.attack = name
        self._attacks = OrderedDict()

        self.set_model(model)
        try:
            self.device = next(model.parameters()).device
        except Exception:
            self.device = None
            print("Failed to set device automatically, please try set_device() manual.")

        # Controls attack mode.
        self.attack_mode = "default"
        self.supported_mode = ["default"]
        self.targeted = False
        self._target_map_function = None

        # Controls when normalization is used.
        self.normalization_used = None
        self._normalization_applied = None
        if self.model.__class__.__name__ == "RobModel":
            self._set_rmodel_normalization_used(model)

        # Controls model mode during attack.
        self._model_training = False
        self._batchnorm_training = False
        self._dropout_training = False

    def forward(self, inputs, labels=None, *args, **kwargs):
        r"""
        It defines the computation performed at every call.
        Should be overridden by all subclasses.
        """
        raise NotImplementedError

    @wrapper_method
    def set_model(self, model):
        self.model = model
        self.model_name = model.__class__.__name__

    def get_logits(self, inputs, labels=None, *args, **kwargs):
        if self._normalization_applied is False:
            inputs = self.normalize(inputs)
        logits = self.model(inputs)
        return logits

    @wrapper_method
    def _set_normalization_applied(self, flag):
        self._normalization_applied = flag

    @wrapper_method
    def set_device(self, device):
        self.device = device

    @wrapper_method
    def _set_rmodel_normalization_used(self, model):
        r"""
        Set attack normalization for MAIR [https://github.com/Harry24k/MAIR].

        """
        mean = getattr(model, "mean", None)
        std = getattr(model, "std", None)
        if (mean is not None) and (std is not None):
            if isinstance(mean, torch.Tensor):
                mean = mean.cpu().numpy()
            if isinstance(std, torch.Tensor):
                std = std.cpu().numpy()
            if (mean != 0).all() or (std != 1).all():
                self.set_normalization_used(mean, std)

    @wrapper_method
    def set_normalization_used(self, mean, std):
        self.normalization_used = {}
        n_channels = len(mean)
        mean = torch.tensor(mean).reshape(1, n_channels, 1, 1)
        std = torch.tensor(std).reshape(1, n_channels, 1, 1)
        self.normalization_used["mean"] = mean
        self.normalization_used["std"] = std
        self._set_normalization_applied(True)

    def normalize(self, inputs):
        mean = self.normalization_used["mean"].to(inputs.device)
        std = self.normalization_used["std"].to(inputs.device)
        return (inputs - mean) / std

    def inverse_normalize(self, inputs):
        mean = self.normalization_used["mean"].to(inputs.device)
        std = self.normalization_used["std"].to(inputs.device)
        return inputs * std + mean

    def get_mode(self):
        r"""
        Get attack mode.

        """
        return self.attack_mode

    @wrapper_method
    def set_mode_default(self):
        r"""
        Set attack mode as default mode.

        """
        self.attack_mode = "default"
        self.targeted = False
        print("Attack mode is changed to 'default.'")

    @wrapper_method
    def _set_mode_targeted(self, mode, quiet):
        if "targeted" not in self.supported_mode:
            raise ValueError("Targeted mode is not supported.")
        self.targeted = True
        self.attack_mode = mode
        if not quiet:
            print("Attack mode is changed to '%s'." % mode)

    @wrapper_method
    def set_mode_targeted_by_function(self, target_map_function, quiet=False):
        r"""
        Set attack mode as targeted.

        Arguments:
            target_map_function (function): Label mapping function.
                e.g. lambda inputs, labels:(labels+1)%10.
                None for using input labels as targeted labels. (Default)
            quiet (bool): Display information message or not. (Default: False)

        """
        self._set_mode_targeted("targeted(custom)", quiet)
        self._target_map_function = target_map_function

    @wrapper_method
    def set_mode_targeted_random(self, quiet=False):
        r"""
        Set attack mode as targeted with random labels.

        Arguments:
            quiet (bool): Display information message or not. (Default: False)

        """
        self._set_mode_targeted("targeted(random)", quiet)
        self._target_map_function = self.get_random_target_label

    @wrapper_method
    def set_mode_targeted_least_likely(self, kth_min=1, quiet=False):
        r"""
        Set attack mode as targeted with least likely labels.

        Arguments:
            kth_min (str): label with the k-th smallest probability used as target labels. (Default: 1)
            num_classses (str): number of classes. (Default: False)

        """
        self._set_mode_targeted("targeted(least-likely)", quiet)
        assert kth_min > 0
        self._kth_min = kth_min
        self._target_map_function = self.get_least_likely_label

    @wrapper_method
    def set_mode_targeted_by_label(self, quiet=False):
        r"""
        Set attack mode as targeted.

        Arguments:
            quiet (bool): Display information message or not. (Default: False)

        .. note::
            Use user-supplied labels as target labels.
        """
        self._set_mode_targeted("targeted(label)", quiet)
        self._target_map_function = "function is a string"

    @wrapper_method
    def set_model_training_mode(
        self, model_training=False, batchnorm_training=False, dropout_training=False
    ):
        r"""
        Set training mode during attack process.

        Arguments:
            model_training (bool): True for using training mode for the entire model during attack process.
            batchnorm_training (bool): True for using training mode for batchnorms during attack process.
            dropout_training (bool): True for using training mode for dropouts during attack process.

        .. note::
            For RNN-based models, we cannot calculate gradients with eval mode.
            Thus, it should be changed to the training mode during the attack.
        """
        self._model_training = model_training
        self._batchnorm_training = batchnorm_training
        self._dropout_training = dropout_training

    @wrapper_method
    def _change_model_mode(self, given_training):
        if self._model_training:
            self.model.train()
            for _, m in self.model.named_modules():
                if not self._batchnorm_training:
                    if "BatchNorm" in m.__class__.__name__:
                        m = m.eval()
                if not self._dropout_training:
                    if "Dropout" in m.__class__.__name__:
                        m = m.eval()
        else:
            self.model.eval()

    @wrapper_method
    def _recover_model_mode(self, given_training):
        if given_training:
            self.model.train()

    def save(
        self,
        data_loader,
        save_path=None,
        verbose=True,
        return_verbose=False,
        save_predictions=False,
        save_clean_inputs=False,
        save_type="float",
    ):
        r"""
        Save adversarial inputs as torch.tensor from given torch.utils.data.DataLoader.

        Arguments:
            save_path (str): save_path.
            data_loader (torch.utils.data.DataLoader): data loader.
            verbose (bool): True for displaying detailed information. (Default: True)
            return_verbose (bool): True for returning detailed information. (Default: False)
            save_predictions (bool): True for saving predicted labels (Default: False)
            save_clean_inputs (bool): True for saving clean inputs (Default: False)

        """
        if save_path is not None:
            adv_input_list = []
            label_list = []
            if save_predictions:
                pred_list = []
            if save_clean_inputs:
                input_list = []

        correct = 0
        total = 0
        l2_distance = []

        total_batch = len(data_loader)
        given_training = self.model.training

        for step, (inputs, labels) in enumerate(data_loader):
            start = time.time()
            adv_inputs = self.__call__(inputs, labels)
            batch_size = len(inputs)

            if verbose or return_verbose:
                with torch.no_grad():
                    outputs = self.get_output_with_eval_nograd(adv_inputs)

                    # Calculate robust accuracy
                    _, pred = torch.max(outputs.data, 1)
                    total += labels.size(0)
                    right_idx = pred == labels.to(self.device)
                    correct += right_idx.sum()
                    rob_acc = 100 * float(correct) / total

                    # Calculate l2 distance
                    delta = (adv_inputs - inputs.to(self.device)).view(
                        batch_size, -1
                    )  # nopep8
                    l2_distance.append(
                        torch.norm(delta[~right_idx], p=2, dim=1)
                    )  # nopep8
                    l2 = torch.cat(l2_distance).mean().item()

                    # Calculate time computation
                    progress = (step + 1) / total_batch * 100
                    end = time.time()
                    elapsed_time = end - start

                    if verbose:
                        self._save_print(
                            progress, rob_acc, l2, elapsed_time, end="\r"
                        )  # nopep8

            if save_path is not None:
                adv_input_list.append(adv_inputs.detach().cpu())
                label_list.append(labels.detach().cpu())

                adv_input_list_cat = torch.cat(adv_input_list, 0)
                label_list_cat = torch.cat(label_list, 0)

                save_dict = {
                    "adv_inputs": adv_input_list_cat,
                    "labels": label_list_cat,
                }  # nopep8

                if save_predictions:
                    pred_list.append(pred.detach().cpu())
                    pred_list_cat = torch.cat(pred_list, 0)
                    save_dict["preds"] = pred_list_cat

                if save_clean_inputs:
                    input_list.append(inputs.detach().cpu())
                    input_list_cat = torch.cat(input_list, 0)
                    save_dict["clean_inputs"] = input_list_cat

                if self.normalization_used is not None:
                    save_dict["adv_inputs"] = self.inverse_normalize(
                        save_dict["adv_inputs"]
                    )  # nopep8
                    if save_clean_inputs:
                        save_dict["clean_inputs"] = self.inverse_normalize(
                            save_dict["clean_inputs"]
                        )  # nopep8

                if save_type == "int":
                    save_dict["adv_inputs"] = self.to_type(
                        save_dict["adv_inputs"], "int"
                    )  # nopep8
                    if save_clean_inputs:
                        save_dict["clean_inputs"] = self.to_type(
                            save_dict["clean_inputs"], "int"
                        )  # nopep8

                save_dict["save_type"] = save_type
                torch.save(save_dict, save_path)

        # To avoid erasing the printed information.
        if verbose:
            self._save_print(progress, rob_acc, l2, elapsed_time, end="\n")

        if given_training:
            self.model.train()

        if return_verbose:
            return rob_acc, l2, elapsed_time

    @staticmethod
    def to_type(inputs, type):
        r"""
        Return inputs as int if float is given.
        """
        if type == "int":
            if isinstance(inputs, torch.FloatTensor) or isinstance(
                inputs, torch.cuda.FloatTensor
            ):
                return (inputs * 255).type(torch.uint8)
        elif type == "float":
            if isinstance(inputs, torch.ByteTensor) or isinstance(
                inputs, torch.cuda.ByteTensor
            ):
                return inputs.float() / 255
        else:
            raise ValueError(type + " is not a valid type. [Options: float, int]")
        return inputs

    @staticmethod
    def _save_print(progress, rob_acc, l2, elapsed_time, end):
        print(
            "- Save progress: %2.2f %% / Robust accuracy: %2.2f %% / L2: %1.5f (%2.3f it/s) \t"
            % (progress, rob_acc, l2, elapsed_time),
            end=end,
        )

    @staticmethod
    def load(
        load_path,
        batch_size=128,
        shuffle=False,
        normalize=None,
        load_predictions=False,
        load_clean_inputs=False,
    ):
        save_dict = torch.load(load_path)
        keys = ["adv_inputs", "labels"]

        if load_predictions:
            keys.append("preds")
        if load_clean_inputs:
            keys.append("clean_inputs")

        if save_dict["save_type"] == "int":
            save_dict["adv_inputs"] = save_dict["adv_inputs"].float() / 255
            if load_clean_inputs:
                save_dict["clean_inputs"] = (
                    save_dict["clean_inputs"].float() / 255
                )  # nopep8

        if normalize is not None:
            n_channels = len(normalize["mean"])
            mean = torch.tensor(normalize["mean"]).reshape(1, n_channels, 1, 1)
            std = torch.tensor(normalize["std"]).reshape(1, n_channels, 1, 1)
            save_dict["adv_inputs"] = (save_dict["adv_inputs"] - mean) / std
            if load_clean_inputs:
                save_dict["clean_inputs"] = (
                    save_dict["clean_inputs"] - mean
                ) / std  # nopep8

        adv_data = TensorDataset(*[save_dict[key] for key in keys])
        adv_loader = DataLoader(adv_data, batch_size=batch_size, shuffle=shuffle)
        print(
            "Data is loaded in the following order: [%s]" % (", ".join(keys))
        )  # nopep8
        return adv_loader

    @torch.no_grad()
    def get_output_with_eval_nograd(self, inputs):
        given_training = self.model.training
        if given_training:
            self.model.eval()
        outputs = self.get_logits(inputs)
        if given_training:
            self.model.train()
        return outputs

    def get_target_label(self, inputs, labels=None):
        r"""
        Function for changing the attack mode.
        Return input labels.
        """
        if self._target_map_function is None:
            raise ValueError(
                "target_map_function is not initialized by set_mode_targeted."
            )
        if self.attack_mode == "targeted(label)":
            target_labels = labels
        else:
            target_labels = self._target_map_function(inputs, labels)
        return target_labels

    @torch.no_grad()
    def get_least_likely_label(self, inputs, labels=None):
        outputs = self.get_output_with_eval_nograd(inputs)
        if labels is None:
            _, labels = torch.max(outputs, dim=1)
        n_classses = outputs.shape[-1]

        target_labels = torch.zeros_like(labels)
        for counter in range(labels.shape[0]):
            l = list(range(n_classses))
            l.remove(labels[counter])
            _, t = torch.kthvalue(outputs[counter][l], self._kth_min)
            target_labels[counter] = l[t]

        return target_labels.long().to(self.device)

    @torch.no_grad()
    def get_random_target_label(self, inputs, labels=None):
        outputs = self.get_output_with_eval_nograd(inputs)
        if labels is None:
            _, labels = torch.max(outputs, dim=1)
        n_classses = outputs.shape[-1]

        target_labels = torch.zeros_like(labels)
        for counter in range(labels.shape[0]):
            l = list(range(n_classses))
            l.remove(labels[counter])
            t = (len(l) * torch.rand([1])).long().to(self.device)
            target_labels[counter] = l[t]

        return target_labels.long().to(self.device)

    def __call__(self, inputs, labels=None, *args, **kwargs):
        given_training = self.model.training
        self._change_model_mode(given_training)

        if self._normalization_applied is True:
            inputs = self.inverse_normalize(inputs)
            self._set_normalization_applied(False)

            adv_inputs = self.forward(inputs, labels, *args, **kwargs)
            # adv_inputs = self.to_type(adv_inputs, self.return_type)

            adv_inputs = self.normalize(adv_inputs)
            self._set_normalization_applied(True)
        else:
            adv_inputs = self.forward(inputs, labels, *args, **kwargs)
            # adv_inputs = self.to_type(adv_inputs, self.return_type)

        self._recover_model_mode(given_training)

        return adv_inputs

    def __repr__(self):
        info = self.__dict__.copy()

        del_keys = ["model", "attack", "supported_mode"]

        for key in info.keys():
            if key[0] == "_":
                del_keys.append(key)

        for key in del_keys:
            del info[key]

        info["attack_mode"] = self.attack_mode
        info["normalization_used"] = (
            True if self.normalization_used is not None else False
        )

        return (
            self.attack
            + "("
            + ", ".join("{}={}".format(key, val) for key, val in info.items())
            + ")"
        )

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

        attacks = self.__dict__.get("_attacks")

        # Get all items in iterable items.
        def get_all_values(items, stack=[]):
            if items not in stack:
                stack.append(items)
                if isinstance(items, list) or isinstance(items, dict):
                    if isinstance(items, dict):
                        items = list(items.keys()) + list(items.values())
                    for item in items:
                        yield from get_all_values(item, stack)
                else:
                    if isinstance(items, Attack):
                        yield items
            else:
                if isinstance(items, Attack):
                    yield items

        for num, value in enumerate(get_all_values(value)):
            attacks[name + "." + str(num)] = value
            for subname, subvalue in value.__dict__.get("_attacks").items():
                attacks[name + "." + subname] = subvalue



import torch


class Cutout:
    def __init__(self, size=16, p=0.5):
        self.size = size
        self.half_size = size // 2
        self.p = p

    def __call__(self, image):
        assert len(image.shape) == 3  # It should be non-batch image.
        if torch.rand([1]).item() > self.p:
            return image

        left = torch.randint(
            -self.half_size, image.size(1) - self.half_size, [1]
        ).item()
        top = torch.randint(-self.half_size, image.size(2) - self.half_size, [1]).item()
        right = min(image.size(1), left + self.size)
        bottom = min(image.size(2), top + self.size)

        image[:, max(0, left) : right, max(0, top) : bottom] = 0
        return image



import torch
import torch.nn as nn

from ..attack import Attack


class FGSM(Attack):
    r"""
    FGSM in the paper 'Explaining and harnessing adversarial examples'
    [https://arxiv.org/abs/1412.6572]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.FGSM(model, eps=8/255)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, eps=8 / 255):
        super().__init__("FGSM", model)
        self.eps = eps
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()

        images.requires_grad = True
        outputs = self.get_logits(images)

        # Calculate loss
        if self.targeted:
            cost = -loss(outputs, target_labels)
        else:
            cost = loss(outputs, labels)

        # Update adversarial images
        grad = torch.autograd.grad(
            cost, images, retain_graph=False, create_graph=False
        )[0]

        adv_images = images + self.eps * grad.sign()
        adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        return adv_images



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
# This is a modified version of https://github.com/pytorch/vision/blob/master/torchvision/datasets/cifar.py

import os
import os.path
import pickle
import sys

import numpy as np
import torch.utils.data as data
from PIL import Image
from torchvision.datasets.utils import check_integrity, download_url


class CIFAR10(data.Dataset):
    """`CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    Args:
        root (string): Root directory of dataset where directory
            ``cifar-10-batches-py`` exists.
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
    """

    base_folder = "cifar-10-batches-py"
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = "c58f30108f718f92721af3b95e74349a"
    train_list = [
        ["data_batch_1", "c99cafc152244af753f735de768cd75f"],
        ["data_batch_2", "d4bba439e000b95fd0a9bffe97cbabec"],
        ["data_batch_3", "54ebc095f3ab1f0389bbae665268c751"],
        ["data_batch_4", "634d18415352ddfa80567beed471001a"],
        ["data_batch_5", "482c414d41f54cd18b22e5b47cb7c3cb"],
    ]

    test_list = [
        ["test_batch", "40351d587109b95175f43aff81a1287e"],
    ]

    def __init__(
        self,
        root,
        train=True,
        transform=None,
        target_transform=None,
        download=False,
        coarse=False,
    ):
        self.root = os.path.expanduser(root)
        self.transform = transform
        self.target_transform = target_transform
        self.train = train  # training set or test set
        self.coarse = coarse

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        # now load the picked numpy arrays
        if self.train:
            self.train_data = []
            self.train_labels = []
            self.train_coarse_labels = []
            for fentry in self.train_list:
                f = fentry[0]
                file = os.path.join(self.root, self.base_folder, f)
                fo = open(file, "rb")
                if sys.version_info[0] == 2:
                    entry = pickle.load(fo)
                else:
                    entry = pickle.load(fo, encoding="latin1")
                self.train_data.append(entry["data"])
                if "labels" in entry:
                    self.train_labels += entry["labels"]
                else:
                    self.train_labels += entry["fine_labels"]
                    if self.coarse:
                        self.train_coarse_labels += entry["coarse_labels"]
                fo.close()

            self.train_data = np.concatenate(self.train_data)
            self.train_data = self.train_data.reshape((50000, 3, 32, 32))
            self.train_data = self.train_data.transpose(
                (0, 2, 3, 1)
            )  # convert to HWC
        else:
            f = self.test_list[0][0]
            file = os.path.join(self.root, self.base_folder, f)
            fo = open(file, "rb")
            if sys.version_info[0] == 2:
                entry = pickle.load(fo)
            else:
                entry = pickle.load(fo, encoding="latin1")
            self.test_data = entry["data"]
            if "labels" in entry:
                self.test_labels = entry["labels"]
            else:
                self.test_labels = entry["fine_labels"]
                if self.coarse:
                    self.test_coarse_labels = entry["coarse_labels"]
            fo.close()
            self.test_data = self.test_data.reshape((10000, 3, 32, 32))
            self.test_data = self.test_data.transpose(
                (0, 2, 3, 1)
            )  # convert to HWC

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        if self.train:
            img, target = self.train_data[index], self.train_labels[index]
            if self.coarse:
                coarse_target = self.train_coarse_labels[index]
        else:
            img, target = self.test_data[index], self.test_labels[index]
            if self.coarse:
                coarse_target = self.test_coarse_labels[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)
        if not self.coarse:
            return img, target
        else:
            return img, target, coarse_target

    def __len__(self):
        if self.train:
            return len(self.train_data)
        else:
            return len(self.test_data)

    def _check_integrity(self):
        root = self.root
        for fentry in self.train_list + self.test_list:
            filename, md5 = fentry[0], fentry[1]
            fpath = os.path.join(root, self.base_folder, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self):
        import tarfile

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        root = self.root
        download_url(self.url, root, self.filename, self.tgz_md5)

        # extract file
        cwd = os.getcwd()
        tar = tarfile.open(os.path.join(root, self.filename), "r:gz")
        os.chdir(root)
        tar.extractall()
        tar.close()
        os.chdir(cwd)


class CIFAR100(CIFAR10):
    """`CIFAR100 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    This is a subclass of the `CIFAR10` Dataset.
    """

    base_folder = "cifar-100-python"
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = "eb9058c3a382ffc7106e4002c42a8d85"
    train_list = [
        ["train", "16019d7e3df5f24257cddd939b257f8d"],
    ]

    test_list = [
        ["test", "f0ef6b0ae62326f3e7ffdfab6717acfc"],
    ]



#############################################################################################
#  Copied from https://github.com/bearpaw/pytorch-classification/blob/master/utils/eval.py  #
#############################################################################################


__all__ = ["accuracy"]


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    # output: [B, d]
    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    # pred: [B, len(topk)] -> [len(topk), B]
    # target.view(1, -1).expand_as(pred): [len(topk), B]
    correct = pred.eq(
        target.reshape(1, -1).expand_as(pred)
    )  # replaced view with reshape for RN18
    # correct: [len(topk), B]

    res = []
    for k in topk:
        correct_k = (
            correct[:k].reshape(-1).float().sum(0)
        )  # replaced view with reshape for RN18
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


if __name__ == "__main__":
    import torch

    outputs = torch.tensor([[0.3, 0.4], [1.0, 0.8], [0.9, 0.6]])
    targets = torch.tensor([1, 0, 1])
    print(accuracy(outputs, targets, topk=(1, 2)))



import torch
import torch.nn as nn

# https://github.com/YisenWang/dynamic_adv_training/blob/master/models.py
class MNIST_DAT(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.layers = nn.Sequential(  # 28
            nn.Conv2d(1, 32, 3, padding=1),  # , 28-3+1+2=28
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.MaxPool2d(2),  # 14
            nn.Conv2d(32, 64, 3, padding=1),  # 14
            nn.BatchNorm2d(64),
            nn.MaxPool2d(2),  # 7
            nn.Flatten(),
            nn.Linear(64 * 7 * 7, 128),
            nn.ReLU(),
            nn.Linear(128, num_classes),
        )

    def forward(self, x):
        x = self.layers(x)
        return x



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time
import math

import torch
import torch.nn.functional as F

# zero_gradients deprecated in torch >= 1.9.
# zero_gradients is re-defined in the bottom of the code.
# from torch.autograd.gradcheck import zero_gradients
from collections import abc as container_abcs

from ..attack import Attack


class FAB(Attack):
    r"""
    Fast Adaptive Boundary Attack in the paper 'Minimally distorted Adversarial Examples with a Fast Adaptive Boundary Attack'
    [https://arxiv.org/abs/1907.02044]
    [https://github.com/fra31/auto-attack]

    Distance Measure : Linf, L2, L1

    Arguments:
        model (nn.Module): model to attack.
        norm (str) : Lp-norm to minimize. ['Linf', 'L2', 'L1'] (Default: 'Linf')
        eps (float): maximum perturbation. (Default: 8/255)
        steps (int): number of steps. (Default: 10)
        n_restarts (int): number of random restarts. (Default: 1)
        alpha_max (float): alpha_max. (Default: 0.1)
        eta (float): overshooting. (Default: 1.05)
        beta (float): backward step. (Default: 0.9)
        verbose (bool): print progress. (Default: False)
        seed (int): random seed for the starting point. (Default: 0)
        targeted (bool): targeted attack for every wrong classes. (Default: False)
        n_classes (int): number of classes. (Default: 10)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.FAB(model, norm='Linf', steps=10, eps=8/255, n_restarts=1, alpha_max=0.1, eta=1.05, beta=0.9, loss_fn=None, verbose=False, seed=0, targeted=False, n_classes=10)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        norm="Linf",
        eps=8 / 255,
        steps=10,
        n_restarts=1,
        alpha_max=0.1,
        eta=1.05,
        beta=0.9,
        verbose=False,
        seed=0,
        multi_targeted=False,
        n_classes=10,
    ):
        super().__init__("FAB", model)
        self.norm = norm
        self.n_restarts = n_restarts
        Default_EPS_DICT_BY_NORM = {"Linf": 0.3, "L2": 1.0, "L1": 5.0}
        self.eps = eps if eps is not None else Default_EPS_DICT_BY_NORM[norm]
        self.alpha_max = alpha_max
        self.eta = eta
        self.beta = beta
        self.steps = steps
        self.verbose = verbose
        self.seed = seed
        self.target_class = None
        self.multi_targeted = multi_targeted
        self.n_target_classes = n_classes - 1
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)
        adv_images = self.perturb(images, labels)

        return adv_images

    def _get_predicted_label(self, x):
        with torch.no_grad():
            outputs = self.get_logits(x)
        _, y = torch.max(outputs, dim=1)
        return y

    def check_shape(self, x):
        return x if len(x.shape) > 0 else x.unsqueeze(0)

    def get_diff_logits_grads_batch(self, imgs, la):
        im = imgs.clone().requires_grad_()
        with torch.enable_grad():
            y = self.get_logits(im)

        g2 = torch.zeros([y.shape[-1], *imgs.size()]).to(self.device)
        grad_mask = torch.zeros_like(y)
        for counter in range(y.shape[-1]):
            zero_gradients(im)
            grad_mask[:, counter] = 1.0
            y.backward(grad_mask, retain_graph=True)
            grad_mask[:, counter] = 0.0
            g2[counter] = im.grad.data

        g2 = torch.transpose(g2, 0, 1).detach()
        # y2 = self.get_logits(imgs).detach()
        y2 = y.detach()
        df = y2 - y2[torch.arange(imgs.shape[0]), la].unsqueeze(1)
        dg = g2 - g2[torch.arange(imgs.shape[0]), la].unsqueeze(1)
        df[torch.arange(imgs.shape[0]), la] = 1e10

        return df, dg

    def get_diff_logits_grads_batch_targeted(self, imgs, la, la_target):
        u = torch.arange(imgs.shape[0])
        im = imgs.clone().requires_grad_()
        with torch.enable_grad():
            y = self.get_logits(im)
            diffy = -(y[u, la] - y[u, la_target])
            sumdiffy = diffy.sum()

        zero_gradients(im)
        sumdiffy.backward()
        graddiffy = im.grad.data
        df = diffy.detach().unsqueeze(1)
        dg = graddiffy.unsqueeze(1)

        return df, dg

    def attack_single_run(self, x, y=None, use_rand_start=False):
        """
        :param x:    clean images
        :param y:    clean labels, if None we use the predicted labels
        """

        # self.device = x.device
        self.orig_dim = list(x.shape[1:])
        self.ndims = len(self.orig_dim)

        x = x.detach().clone().float().to(self.device)
        # assert next(self.model.parameters()).device == x.device

        y_pred = self._get_predicted_label(x)
        if y is None:
            y = y_pred.detach().clone().long().to(self.device)
        else:
            y = y.detach().clone().long().to(self.device)
        pred = y_pred == y
        corr_classified = pred.float().sum()
        if self.verbose:
            print("Clean accuracy: {:.2%}".format(pred.float().mean()))
        if pred.sum() == 0:
            return x
        pred = self.check_shape(pred.nonzero().squeeze())

        startt = time.time()
        # runs the attack only on correctly classified points
        im2 = x[pred].detach().clone()
        la2 = y[pred].detach().clone()
        if len(im2.shape) == self.ndims:
            im2 = im2.unsqueeze(0)
        bs = im2.shape[0]
        u1 = torch.arange(bs)
        adv = im2.clone()
        adv_c = x.clone()
        res2 = 1e10 * torch.ones([bs]).to(self.device)
        res_c = torch.zeros([x.shape[0]]).to(self.device)
        x1 = im2.clone()
        x0 = im2.clone().reshape([bs, -1])
        counter_restarts = 0

        while counter_restarts < 1:
            if use_rand_start:
                if self.norm == "Linf":
                    t = 2 * torch.rand(x1.shape).to(self.device) - 1
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            t.reshape([t.shape[0], -1])
                            .abs()
                            .max(dim=1, keepdim=True)[0]
                            .reshape([-1, *[1] * self.ndims])
                        )
                        * 0.5
                    )
                elif self.norm == "L2":
                    t = torch.randn(x1.shape).to(self.device)
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            (t ** 2)
                            .view(t.shape[0], -1)
                            .sum(dim=-1)
                            .sqrt()
                            .view(t.shape[0], *[1] * self.ndims)
                        )
                        * 0.5
                    )
                elif self.norm == "L1":
                    t = torch.randn(x1.shape).to(self.device)
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            t.abs()
                            .view(t.shape[0], -1)
                            .sum(dim=-1)
                            .view(t.shape[0], *[1] * self.ndims)
                        )
                        / 2
                    )

                x1 = x1.clamp(0.0, 1.0)

            counter_iter = 0
            while counter_iter < self.steps:
                with torch.no_grad():
                    df, dg = self.get_diff_logits_grads_batch(x1, la2)
                    if self.norm == "Linf":
                        dist1 = df.abs() / (
                            1e-12
                            + dg.abs().view(dg.shape[0], dg.shape[1], -1).sum(dim=-1)
                        )
                    elif self.norm == "L2":
                        dist1 = df.abs() / (
                            1e-12
                            + (dg ** 2)
                            .view(dg.shape[0], dg.shape[1], -1)
                            .sum(dim=-1)
                            .sqrt()
                        )
                    elif self.norm == "L1":
                        dist1 = df.abs() / (
                            1e-12
                            + dg.abs()
                            .reshape([df.shape[0], df.shape[1], -1])
                            .max(dim=2)[0]
                        )
                    else:
                        raise ValueError("norm not supported")
                    ind = dist1.min(dim=1)[1]
                    dg2 = dg[u1, ind]
                    b = -df[u1, ind] + (dg2 * x1).view(x1.shape[0], -1).sum(dim=-1)
                    w = dg2.reshape([bs, -1])

                    if self.norm == "Linf":
                        d3 = projection_linf(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    elif self.norm == "L2":
                        d3 = projection_l2(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    elif self.norm == "L1":
                        d3 = projection_l1(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    d1 = torch.reshape(d3[:bs], x1.shape)
                    d2 = torch.reshape(d3[-bs:], x1.shape)
                    if self.norm == "Linf":
                        a0 = (
                            d3.abs()
                            .max(dim=1, keepdim=True)[0]
                            .view(-1, *[1] * self.ndims)
                        )
                    elif self.norm == "L2":
                        a0 = (
                            (d3 ** 2)
                            .sum(dim=1, keepdim=True)
                            .sqrt()
                            .view(-1, *[1] * self.ndims)
                        )
                    elif self.norm == "L1":
                        a0 = (
                            d3.abs()
                            .sum(dim=1, keepdim=True)
                            .view(-1, *[1] * self.ndims)
                        )
                    a0 = torch.max(a0, 1e-8 * torch.ones(a0.shape).to(self.device))
                    a1 = a0[:bs]
                    a2 = a0[-bs:]
                    alpha = torch.min(
                        torch.max(
                            a1 / (a1 + a2), torch.zeros(a1.shape).to(self.device)
                        ),
                        self.alpha_max * torch.ones(a1.shape).to(self.device),
                    )
                    x1 = (
                        (x1 + self.eta * d1) * (1 - alpha)
                        + (im2 + d2 * self.eta) * alpha
                    ).clamp(0.0, 1.0)

                    is_adv = self._get_predicted_label(x1) != la2

                    if is_adv.sum() > 0:
                        ind_adv = is_adv.nonzero().squeeze()
                        ind_adv = self.check_shape(ind_adv)
                        if self.norm == "Linf":
                            t = (
                                (x1[ind_adv] - im2[ind_adv])
                                .reshape([ind_adv.shape[0], -1])
                                .abs()
                                .max(dim=1)[0]
                            )
                        elif self.norm == "L2":
                            t = (
                                ((x1[ind_adv] - im2[ind_adv]) ** 2)
                                .view(ind_adv.shape[0], -1)
                                .sum(dim=-1)
                                .sqrt()
                            )
                        elif self.norm == "L1":
                            t = (
                                (x1[ind_adv] - im2[ind_adv])
                                .abs()
                                .view(ind_adv.shape[0], -1)
                                .sum(dim=-1)
                            )
                        adv[ind_adv] = x1[ind_adv] * (
                            t < res2[ind_adv]
                        ).float().reshape([-1, *[1] * self.ndims]) + adv[ind_adv] * (
                            t >= res2[ind_adv]
                        ).float().reshape(
                            [-1, *[1] * self.ndims]
                        )
                        res2[ind_adv] = (
                            t * (t < res2[ind_adv]).float()
                            + res2[ind_adv] * (t >= res2[ind_adv]).float()
                        )
                        x1[ind_adv] = (
                            im2[ind_adv] + (x1[ind_adv] - im2[ind_adv]) * self.beta
                        )

                    counter_iter += 1

            counter_restarts += 1

        ind_succ = res2 < 1e10
        if self.verbose:
            print(
                "success rate: {:.0f}/{:.0f}".format(
                    ind_succ.float().sum(), corr_classified
                )
                + " (on correctly classified points) in {:.1f} s".format(
                    time.time() - startt
                )
            )

        res_c[pred] = res2 * ind_succ.float() + 1e10 * (1 - ind_succ.float())
        ind_succ = self.check_shape(ind_succ.nonzero().squeeze())
        adv_c[pred[ind_succ]] = adv[ind_succ].clone()

        return adv_c

    def attack_single_run_targeted(self, x, y=None, use_rand_start=False):
        """
        :param x:    clean images
        :param y:    clean labels, if None we use the predicted labels
        """

        if self.device is None:
            self.device = x.device
        self.orig_dim = list(x.shape[1:])
        self.ndims = len(self.orig_dim)

        x = x.detach().clone().float().to(self.device)
        # assert next(self.model.parameters()).device == x.device

        y_pred = self._get_predicted_label(x)
        if y is None:
            y = y_pred.detach().clone().long().to(self.device)
        else:
            y = y.detach().clone().long().to(self.device)
        pred = y_pred == y
        corr_classified = pred.float().sum()
        if self.verbose:
            print("Clean accuracy: {:.2%}".format(pred.float().mean()))
        if pred.sum() == 0:
            return x
        pred = self.check_shape(pred.nonzero().squeeze())

        output = self.get_logits(x)
        if self.multi_targeted:
            la_target = output.sort(dim=-1)[1][:, -self.target_class]
        else:
            la_target = self.target_class

        startt = time.time()
        # runs the attack only on correctly classified points
        im2 = x[pred].detach().clone()
        la2 = y[pred].detach().clone()
        la_target2 = la_target[pred].detach().clone()
        if len(im2.shape) == self.ndims:
            im2 = im2.unsqueeze(0)
        bs = im2.shape[0]
        u1 = torch.arange(bs)
        adv = im2.clone()
        adv_c = x.clone()
        res2 = 1e10 * torch.ones([bs]).to(self.device)
        res_c = torch.zeros([x.shape[0]]).to(self.device)
        x1 = im2.clone()
        x0 = im2.clone().reshape([bs, -1])
        counter_restarts = 0

        while counter_restarts < 1:
            if use_rand_start:
                if self.norm == "Linf":
                    t = 2 * torch.rand(x1.shape).to(self.device) - 1
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            t.reshape([t.shape[0], -1])
                            .abs()
                            .max(dim=1, keepdim=True)[0]
                            .reshape([-1, *[1] * self.ndims])
                        )
                        * 0.5
                    )
                elif self.norm == "L2":
                    t = torch.randn(x1.shape).to(self.device)
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            (t ** 2)
                            .view(t.shape[0], -1)
                            .sum(dim=-1)
                            .sqrt()
                            .view(t.shape[0], *[1] * self.ndims)
                        )
                        * 0.5
                    )
                elif self.norm == "L1":
                    t = torch.randn(x1.shape).to(self.device)
                    x1 = (
                        im2
                        + (
                            torch.min(
                                res2, self.eps * torch.ones(res2.shape).to(self.device)
                            ).reshape([-1, *[1] * self.ndims])
                        )
                        * t
                        / (
                            t.abs()
                            .view(t.shape[0], -1)
                            .sum(dim=-1)
                            .view(t.shape[0], *[1] * self.ndims)
                        )
                        / 2
                    )

                x1 = x1.clamp(0.0, 1.0)

            counter_iter = 0
            while counter_iter < self.steps:
                with torch.no_grad():
                    df, dg = self.get_diff_logits_grads_batch_targeted(
                        x1, la2, la_target2
                    )
                    if self.norm == "Linf":
                        dist1 = df.abs() / (
                            1e-12
                            + dg.abs().view(dg.shape[0], dg.shape[1], -1).sum(dim=-1)
                        )
                    elif self.norm == "L2":
                        dist1 = df.abs() / (
                            1e-12
                            + (dg ** 2)
                            .view(dg.shape[0], dg.shape[1], -1)
                            .sum(dim=-1)
                            .sqrt()
                        )
                    elif self.norm == "L1":
                        dist1 = df.abs() / (
                            1e-12
                            + dg.abs()
                            .reshape([df.shape[0], df.shape[1], -1])
                            .max(dim=2)[0]
                        )
                    else:
                        raise ValueError("norm not supported")
                    ind = dist1.min(dim=1)[1]

                    dg2 = dg[u1, ind]
                    b = -df[u1, ind] + (dg2 * x1).view(x1.shape[0], -1).sum(dim=-1)
                    w = dg2.reshape([bs, -1])

                    if self.norm == "Linf":
                        d3 = projection_linf(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    elif self.norm == "L2":
                        d3 = projection_l2(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    elif self.norm == "L1":
                        d3 = projection_l1(
                            torch.cat((x1.reshape([bs, -1]), x0), 0),
                            torch.cat((w, w), 0),
                            torch.cat((b, b), 0),
                        )
                    d1 = torch.reshape(d3[:bs], x1.shape)
                    d2 = torch.reshape(d3[-bs:], x1.shape)
                    if self.norm == "Linf":
                        a0 = (
                            d3.abs()
                            .max(dim=1, keepdim=True)[0]
                            .view(-1, *[1] * self.ndims)
                        )
                    elif self.norm == "L2":
                        a0 = (
                            (d3 ** 2)
                            .sum(dim=1, keepdim=True)
                            .sqrt()
                            .view(-1, *[1] * self.ndims)
                        )
                    elif self.norm == "L1":
                        a0 = (
                            d3.abs()
                            .sum(dim=1, keepdim=True)
                            .view(-1, *[1] * self.ndims)
                        )
                    a0 = torch.max(a0, 1e-8 * torch.ones(a0.shape).to(self.device))
                    a1 = a0[:bs]
                    a2 = a0[-bs:]
                    alpha = torch.min(
                        torch.max(
                            a1 / (a1 + a2), torch.zeros(a1.shape).to(self.device)
                        ),
                        self.alpha_max * torch.ones(a1.shape).to(self.device),
                    )
                    x1 = (
                        (x1 + self.eta * d1) * (1 - alpha)
                        + (im2 + d2 * self.eta) * alpha
                    ).clamp(0.0, 1.0)

                    is_adv = self._get_predicted_label(x1) != la2

                    if is_adv.sum() > 0:
                        ind_adv = is_adv.nonzero().squeeze()
                        ind_adv = self.check_shape(ind_adv)
                        if self.norm == "Linf":
                            t = (
                                (x1[ind_adv] - im2[ind_adv])
                                .reshape([ind_adv.shape[0], -1])
                                .abs()
                                .max(dim=1)[0]
                            )
                        elif self.norm == "L2":
                            t = (
                                ((x1[ind_adv] - im2[ind_adv]) ** 2)
                                .view(ind_adv.shape[0], -1)
                                .sum(dim=-1)
                                .sqrt()
                            )
                        elif self.norm == "L1":
                            t = (
                                (x1[ind_adv] - im2[ind_adv])
                                .abs()
                                .view(ind_adv.shape[0], -1)
                                .sum(dim=-1)
                            )
                        adv[ind_adv] = x1[ind_adv] * (
                            t < res2[ind_adv]
                        ).float().reshape([-1, *[1] * self.ndims]) + adv[ind_adv] * (
                            t >= res2[ind_adv]
                        ).float().reshape(
                            [-1, *[1] * self.ndims]
                        )
                        res2[ind_adv] = (
                            t * (t < res2[ind_adv]).float()
                            + res2[ind_adv] * (t >= res2[ind_adv]).float()
                        )
                        x1[ind_adv] = (
                            im2[ind_adv] + (x1[ind_adv] - im2[ind_adv]) * self.beta
                        )

                    counter_iter += 1

            counter_restarts += 1

        ind_succ = res2 < 1e10
        if self.verbose:
            print(
                "success rate: {:.0f}/{:.0f}".format(
                    ind_succ.float().sum(), corr_classified
                )
                + " (on correctly classified points) in {:.1f} s".format(
                    time.time() - startt
                )
            )

        res_c[pred] = res2 * ind_succ.float() + 1e10 * (1 - ind_succ.float())
        ind_succ = self.check_shape(ind_succ.nonzero().squeeze())
        adv_c[pred[ind_succ]] = adv[ind_succ].clone()

        return adv_c

    def perturb(self, x, y):
        adv = x.clone()
        with torch.no_grad():
            acc = self.get_logits(x).max(1)[1] == y

            startt = time.time()

            torch.random.manual_seed(self.seed)
            torch.cuda.random.manual_seed(self.seed)

            def inner_perturb(targeted):
                for counter in range(self.n_restarts):
                    ind_to_fool = acc.nonzero().squeeze()
                    if len(ind_to_fool.shape) == 0:
                        ind_to_fool = ind_to_fool.unsqueeze(0)
                    if ind_to_fool.numel() != 0:
                        x_to_fool, y_to_fool = (
                            x[ind_to_fool].clone(),
                            y[ind_to_fool].clone(),
                        )  # nopep8

                        if targeted:
                            adv_curr = self.attack_single_run_targeted(
                                x_to_fool, y_to_fool, use_rand_start=(counter > 0)
                            )
                        else:
                            adv_curr = self.attack_single_run(
                                x_to_fool, y_to_fool, use_rand_start=(counter > 0)
                            )

                        acc_curr = self.get_logits(adv_curr).max(1)[1] == y_to_fool
                        if self.norm == "Linf":
                            res = (
                                (x_to_fool - adv_curr)
                                .abs()
                                .view(x_to_fool.shape[0], -1)
                                .max(1)[0]
                            )  # nopep8
                        elif self.norm == "L2":
                            res = (
                                ((x_to_fool - adv_curr) ** 2)
                                .view(x_to_fool.shape[0], -1)
                                .sum(dim=-1)
                                .sqrt()
                            )  # nopep8
                        acc_curr = torch.max(acc_curr, res > self.eps)

                        ind_curr = (acc_curr == 0).nonzero().squeeze()
                        acc[ind_to_fool[ind_curr]] = 0
                        adv[ind_to_fool[ind_curr]] = adv_curr[ind_curr].clone()

                        if self.verbose:
                            if targeted:
                                print(
                                    "restart {} - target_class {} - robust accuracy: {:.2%} at eps = {:.5f} - cum. time: {:.1f} s".format(
                                        counter,
                                        self.target_class,
                                        acc.float().mean(),
                                        self.eps,
                                        time.time() - startt,
                                    )
                                )
                            else:
                                print(
                                    "restart {} - robust accuracy: {:.2%} at eps = {:.5f} - cum. time: {:.1f} s".format(
                                        counter,
                                        acc.float().mean(),
                                        self.eps,
                                        time.time() - startt,
                                    )
                                )

            if self.multi_targeted:
                for target_class in range(2, self.n_target_classes + 2):
                    self.target_class = target_class
                    inner_perturb(targeted=True)
            elif self.targeted:
                self.target_class = self.get_target_label(x, y)
                inner_perturb(targeted=True)
            else:
                inner_perturb(targeted=False)
        return adv


def projection_linf(points_to_project, w_hyperplane, b_hyperplane):
    device = points_to_project.device
    t, w, b = points_to_project, w_hyperplane.clone(), b_hyperplane.clone()

    sign = 2 * ((w * t).sum(1) - b >= 0) - 1
    w.mul_(sign.unsqueeze(1))
    b.mul_(sign)

    a = (w < 0).float()
    d = (a - t) * (w != 0).float()

    p = a - t * (2 * a - 1)
    indp = torch.argsort(p, dim=1)

    b = b - (w * t).sum(1)
    b0 = (w * d).sum(1)

    indp2 = indp.flip((1,))
    ws = w.gather(1, indp2)
    bs2 = -ws * d.gather(1, indp2)

    s = torch.cumsum(ws.abs(), dim=1)
    sb = torch.cumsum(bs2, dim=1) + b0.unsqueeze(1)

    b2 = sb[:, -1] - s[:, -1] * p.gather(1, indp[:, 0:1]).squeeze(1)
    c_l = b - b2 > 0
    c2 = (b - b0 > 0) & (~c_l)
    lb = torch.zeros(c2.sum(), device=device)
    ub = torch.full_like(lb, w.shape[1] - 1)
    nitermax = math.ceil(math.log2(w.shape[1]))

    indp_, sb_, s_, p_, b_ = indp[c2], sb[c2], s[c2], p[c2], b[c2]
    for counter in range(nitermax):
        counter4 = torch.floor((lb + ub) / 2)

        counter2 = counter4.long().unsqueeze(1)
        indcurr = indp_.gather(1, indp_.size(1) - 1 - counter2)
        b2 = (
            sb_.gather(1, counter2) - s_.gather(1, counter2) * p_.gather(1, indcurr)
        ).squeeze(
            1
        )  # nopep8
        c = b_ - b2 > 0

        lb = torch.where(c, counter4, lb)
        ub = torch.where(c, ub, counter4)

    lb = lb.long()

    if c_l.any():
        lmbd_opt = torch.clamp_min(
            (b[c_l] - sb[c_l, -1]) / (-s[c_l, -1]), min=0
        ).unsqueeze(-1)
        d[c_l] = (2 * a[c_l] - 1) * lmbd_opt

    lmbd_opt = torch.clamp_min((b[c2] - sb[c2, lb]) / (-s[c2, lb]), min=0).unsqueeze(-1)
    d[c2] = torch.min(lmbd_opt, d[c2]) * a[c2] + torch.max(-lmbd_opt, d[c2]) * (
        1 - a[c2]
    )

    return d * (w != 0).float()


def projection_l2(points_to_project, w_hyperplane, b_hyperplane):
    device = points_to_project.device
    t, w, b = points_to_project, w_hyperplane.clone(), b_hyperplane

    c = (w * t).sum(1) - b
    ind2 = 2 * (c >= 0) - 1
    w.mul_(ind2.unsqueeze(1))
    c.mul_(ind2)

    r = torch.max(t / w, (t - 1) / w).clamp(min=-1e12, max=1e12)
    r.masked_fill_(w.abs() < 1e-8, 1e12)
    r[r == -1e12] *= -1
    rs, indr = torch.sort(r, dim=1)
    rs2 = F.pad(rs[:, 1:], (0, 1))
    rs.masked_fill_(rs == 1e12, 0)
    rs2.masked_fill_(rs2 == 1e12, 0)

    w3s = (w ** 2).gather(1, indr)
    w5 = w3s.sum(dim=1, keepdim=True)
    ws = w5 - torch.cumsum(w3s, dim=1)
    d = -(r * w)
    d.mul_((w.abs() > 1e-8).float())
    s = torch.cat(
        (-w5 * rs[:, 0:1], torch.cumsum((-rs2 + rs) * ws, dim=1) - w5 * rs[:, 0:1]), 1
    )

    c4 = s[:, 0] + c < 0
    c3 = (d * w).sum(dim=1) + c > 0
    c2 = ~(c4 | c3)

    lb = torch.zeros(c2.sum(), device=device)
    ub = torch.full_like(lb, w.shape[1] - 1)
    nitermax = math.ceil(math.log2(w.shape[1]))

    s_, c_ = s[c2], c[c2]
    for counter in range(nitermax):
        counter4 = torch.floor((lb + ub) / 2)
        counter2 = counter4.long().unsqueeze(1)
        c3 = s_.gather(1, counter2).squeeze(1) + c_ > 0
        lb = torch.where(c3, counter4, lb)
        ub = torch.where(c3, ub, counter4)

    lb = lb.long()

    if c4.any():
        alpha = c[c4] / w5[c4].squeeze(-1)
        d[c4] = -alpha.unsqueeze(-1) * w[c4]

    if c2.any():
        alpha = (s[c2, lb] + c[c2]) / ws[c2, lb] + rs[c2, lb]
        alpha[ws[c2, lb] == 0] = 0
        c5 = (alpha.unsqueeze(-1) > r[c2]).float()
        d[c2] = d[c2] * c5 - alpha.unsqueeze(-1) * w[c2] * (1 - c5)

    return d * (w.abs() > 1e-8).float()


def projection_l1(points_to_project, w_hyperplane, b_hyperplane):
    device = points_to_project.device
    t, w, b = points_to_project, w_hyperplane.clone(), b_hyperplane

    c = (w * t).sum(1) - b
    ind2 = 2 * (c >= 0) - 1
    w.mul_(ind2.unsqueeze(1))
    c.mul_(ind2)

    r = (1 / w).abs().clamp_max(1e12)
    indr = torch.argsort(r, dim=1)
    indr_rev = torch.argsort(indr)

    c6 = (w < 0).float()
    d = (-t + c6) * (w != 0).float()
    ds = torch.min(-w * t, w * (1 - t)).gather(1, indr)
    ds2 = torch.cat((c.unsqueeze(-1), ds), 1)
    s = torch.cumsum(ds2, dim=1)

    c2 = s[:, -1] < 0

    lb = torch.zeros(c2.sum(), device=device)
    ub = torch.full_like(lb, s.shape[1])
    nitermax = math.ceil(math.log2(w.shape[1]))

    s_ = s[c2]
    for counter in range(nitermax):
        counter4 = torch.floor((lb + ub) / 2)
        counter2 = counter4.long().unsqueeze(1)
        c3 = s_.gather(1, counter2).squeeze(1) > 0
        lb = torch.where(c3, counter4, lb)
        ub = torch.where(c3, ub, counter4)

    lb2 = lb.long()

    if c2.any():
        indr = indr[c2].gather(1, lb2.unsqueeze(1)).squeeze(1)
        u = torch.arange(0, w.shape[0], device=device).unsqueeze(1)
        u2 = torch.arange(0, w.shape[1], device=device, dtype=torch.float).unsqueeze(0)
        alpha = -s[c2, lb2] / w[c2, indr]
        c5 = u2 < lb.unsqueeze(-1)
        u3 = c5[u[: c5.shape[0]], indr_rev[c2]]
        d[c2] = d[c2] * u3.float()
        d[c2, indr] = alpha

    return d * (w.abs() > 1e-8).float()


def zero_gradients(x):
    if isinstance(x, torch.Tensor):
        if x.grad is not None:
            x.grad.detach_()
            x.grad.zero_()
    elif isinstance(x, container_abcs.Iterable):
        for elem in x:
            zero_gradients(elem)



import os
import logging
from collections import OrderedDict

import torch
from torch.optim import *
from torch.optim.lr_scheduler import *

from .rm import RecordManager
from .. import RobModel
from ..optim import *


r"""
Base class for all trainers.

"""


class Trainer:
    def __init__(self, rmodel, device=None):
        assert isinstance(rmodel, RobModel)
        self.rmodel = rmodel
        if device is None:
            self.device = next(rmodel.parameters()).device
        else:
            device = device

        # Init train info and record manager
        self.accumulated_epoch = 0
        self.accumulated_iter = 0
        self.curr_epoch = 0
        self.curr_iter = 0
        self.rm = RecordManager()

        # Init dicts for record and save for each iteration
        self.dict_record = OrderedDict()
        self.dict_save = OrderedDict()

        # Init setups
        self.optimizer = None
        self.scheduler = None
        self.scheduler_type = None
        self.minimizer = None
        self.clip_grad_norm = None

    def setup(
        self,
        optimizer,
        scheduler=None,
        scheduler_type=None,
        n_epochs=None,
        n_iters=None,
        minimizer=None,
        clip_grad_norm=None,
    ):
        params = self.rmodel.parameters()
        self.optimizer = self.generate_optimizer(optimizer, params)
        self.scheduler, self.scheduler_type = self.generate_schdeuler(
            scheduler, self.optimizer, scheduler_type, n_epochs, n_iters
        )
        self.minimizer = self.generate_minimizer(minimizer, self.rmodel, self.optimizer)
        self.clip_grad_norm = clip_grad_norm

    def fit(
        self,
        train_loader,
        n_epochs,
        n_iters=None,
        record_type="Epoch",
        save_path=None,
        save_type="Epoch",
        save_best=None,
        save_overwrite=False,
        refit=False,
    ):

        # Check Save and Record Values
        self._check_valid_options(save_type)
        self._check_valid_options(record_type)
        self._check_valid_save_path(save_path, save_type, save_overwrite)

        if refit:
            if self.rm.count == 0:
                raise ValueError("Please call load_dict for refitting.")
            # Update record and init
            self.rm.update(
                record_type=record_type, save_path=save_path, best_option=save_best
            )
            record_type = self.rm.record_type

            self.accumulated_epoch += -1
            start_epoch = self.curr_epoch - 1
            start_iter = self.curr_iter
        else:
            # Start record and save init
            self.rm.initialize(
                record_type=record_type, save_path=save_path, best_option=save_best
            )
            if (save_path is not None) and (self.accumulated_iter == 0):
                self.save_dict(save_path, is_init=True)

            start_epoch = 0
            start_iter = 0

        # Print train information
        self.rm.print(record_type, "[%s]" % self.__class__.__name__)
        self.rm.print(record_type, "Training Information.")
        self.rm.print(record_type, "-Epochs: %s" % n_epochs)
        self.rm.print(record_type, "-Optimizer: %s" % self.optimizer)
        self.rm.print(record_type, "-Scheduler: %s" % self.scheduler)
        self.rm.print(record_type, "-Minmizer: %s" % self.minimizer)
        self.rm.print(record_type, "-Save Path: %s" % save_path)
        self.rm.print(record_type, "-Save Type: %s" % str(save_type))
        self.rm.print(record_type, "-Record Type: %s" % str(record_type))
        self.rm.print(record_type, "-Device: %s" % self.device)

        # Start training
        for epoch in range(start_epoch, n_epochs):
            # Update current epoch and n_iters
            self.curr_epoch = epoch + 1
            self.accumulated_epoch += 1
            if n_iters is None:
                n_iters = len(train_loader)

            for i, train_data in enumerate(train_loader):
                # Update current iteration
                self.curr_iter = i + 1
                if self.curr_iter <= start_iter:  # For refit
                    continue
                self.accumulated_iter += 1
                is_last_batch = self.curr_iter == n_iters

                # Init records and dicts
                self._init_dicts_record_save()
                self.add_record_item("Epoch", self.accumulated_epoch)
                self.add_record_item("Iter", self.curr_iter)

                # Set train mode
                self.rmodel.train()

                # Update weight
                self.rm.progress_start()
                self._update_weight(train_data)
                self.rm.progress_end()

                # Eval mode
                if self._check_run_condition(record_type, is_last_batch):
                    if type(self).record_during_eval != self.record_during_eval:
                        self.rmodel.eval()
                        self.record_during_eval()
                    self.add_record_item("lr", self.optimizer.param_groups[0]["lr"])
                    self.rm.add(self.dict_record)
                    # If record added, save dicts.

                # Update scheduler
                if self._check_run_condition(self.scheduler_type, is_last_batch):
                    self.scheduler.step()

                # Save dicts
                if save_path is not None:
                    is_best = self.rm.check_best(self.dict_record)
                    is_save_condition = self._check_run_condition(
                        save_type, is_last_batch
                    )
                    # Save if condition is satisfied or best or end of the epoch.
                    if is_save_condition or is_best or is_last_batch:
                        self.save_dict(save_path, save_type, is_best=is_best)

                # Check number of iterations
                if is_last_batch:
                    break

            start_iter = 0  # For refit

        if (save_path is not None) and (record_type is not None):
            self.rm.generate_summary()

    def calculate_cost(self, *inputs):
        raise NotImplementedError

    def record_during_eval(self, *args):
        raise NotImplementedError

    def _init_dicts_record_save(self):
        del self.dict_record, self.dict_save
        self.dict_record = OrderedDict()
        self.dict_save = OrderedDict()

    def add_record_item(self, key, val):
        if key in self.dict_record.keys():
            n = 1
            while key + "_" + str(n) in self.dict_record.keys():
                n = n + 1
            self.dict_record[key + "_" + str(n)] = val
        else:
            self.dict_record[key] = val

    def add_save_item(self, key, val):
        if key in self.dict_save.keys():
            n = 1
            while key + "_" + str(n) in self.dict_save.keys():
                n = n + 1
            self.dict_save[key + "_" + str(n)] = val
        else:
            self.dict_save[key] = val

    @staticmethod
    def _check_run_condition(run_type, is_last_batch):
        if run_type == "Epoch" and is_last_batch:
            return True
        elif run_type == "Iter":
            return True
        return False

    @staticmethod
    def generate_optimizer(optimizer, params):
        if isinstance(optimizer, str):
            optimizer = eval(
                optimizer.split("(")[0] + "(params," + optimizer.split("(")[1]
            )
        return optimizer

    @staticmethod
    def generate_schdeuler(
        scheduler, optimizer, scheduler_type=None, n_epochs=None, n_iters=None
    ):
        if isinstance(scheduler, str):
            # Step(milestones=[2, 4], gamma=0.1)
            if scheduler.startswith("Step("):
                scheduler = eval("MultiStepLR(optimizer, " + scheduler.split("(")[1])
                scheduler_type = "Epoch"
            # Cyclic(base_lr=0, max_lr=0.3)
            elif scheduler.startswith("Cyclic("):
                lr_steps = n_epochs * n_iters
                scheduler = eval(
                    "CyclicLR(optimizer, "
                    + scheduler.split("(")[1].split(")")[0]
                    + ", step_size_up=lr_steps/2, step_size_down=lr_steps/2)"
                )
                scheduler_type = "Iter"
            # Cosine
            elif "Cosine" == scheduler:
                scheduler = CosineAnnealingLR(optimizer, n_epochs, eta_min=0)
                scheduler_type = "Epoch"
            else:
                scheduler = eval(
                    scheduler.split("(")[0] + "(optimizer, " + scheduler.split("(")[1]
                )

        if (scheduler is not None) and (scheduler_type is None):
            raise ValueError(
                "The type of scheduler must be specified as 'Epoch' or 'Iter'."
            )

        return scheduler, scheduler_type

    @staticmethod
    def generate_minimizer(minimizer, model, optimizer):
        if isinstance(minimizer, str):
            minimizer = eval(
                minimizer.split("(")[0] + "(model, optimizer," + minimizer.split("(")[1]
            )
        return minimizer

    def _update_weight(self, *inputs):
        if self.minimizer is not None:
            self.minimizer.step(lambda x: self.calculate_cost(x), *inputs)
        else:
            cost = self.calculate_cost(*inputs)
            self.optimizer.zero_grad()
            cost.backward()
            if self.clip_grad_norm:
                torch.nn.utils.clip_grad_norm_(
                    self.rmodel.parameters(), self.clip_grad_norm
                )
            self.optimizer.step()

    @staticmethod
    def _check_valid_options(key):
        if key in ["Epoch", "Iter", None]:
            pass
        else:
            raise ValueError(key, " is not valid. [Hint:'Epoch', 'Iter', None]")

    # Check and Create Path
    @staticmethod
    def _check_and_create_path(path, overwrite=False, file=False):
        if os.path.exists(path):
            if overwrite:
                logging.warning("Save file(s) will be overwritten:" + path)
            else:
                raise ValueError("[%s] is already exists." % (path))
        else:
            if not file:
                os.makedirs(path)

    def _check_valid_save_path(self, save_path, save_type, save_overwrite):
        if save_path is not None:
            if save_path[-1] == "/":
                save_path = save_path[:-1]
            # Save Initial Model
            self._check_and_create_path(save_path, overwrite=save_overwrite)
            self._check_and_create_path(
                save_path + "/init.pth", overwrite=save_overwrite, file=True
            )
            self._check_and_create_path(
                save_path + "/last.pth", overwrite=save_overwrite, file=True
            )
            self._check_and_create_path(
                save_path + "/best.pth", overwrite=save_overwrite, file=True
            )
            self._check_and_create_path(
                save_path + "/log.txt", overwrite=save_overwrite, file=True
            )

            if save_type in ["Epoch", "Iter"]:
                self._check_and_create_path(
                    save_path + "/epoch_iter/", overwrite=save_overwrite
                )
                self.save_dict(save_path, 0)
        else:
            if save_type is not None:
                raise ValueError("save_path should be given for save_type != None.")

    def save_dict(self, save_path, save_type=None, is_init=False, is_best=False):
        self.add_save_item("accumulated_epoch", self.accumulated_epoch)
        self.add_save_item("accumulated_iter", self.accumulated_iter)
        self.add_save_item("curr_epoch", self.curr_epoch)
        self.add_save_item("curr_iter", self.curr_iter)

        self.add_save_item("rmodel", self.rmodel.state_dict())
        if self.optimizer is not None:
            self.add_save_item("optimizer", self.optimizer.state_dict())
        if self.scheduler is not None:
            self.add_save_item("scheduler", self.scheduler.state_dict())
            self.add_save_item("scheduler_type", self.scheduler_type)
        if self.minimizer is not None:
            self.add_save_item("minimizer", self.minimizer.state_dict())
        if self.rm is not None:
            self.add_save_item("recordmanager", self.rm)
        if self.dict_record is not None:
            self.add_save_item("record_info", self.dict_record)

        # Save last and best
        torch.save(self.dict_save, save_path + "/last.pth")
        if is_best:
            torch.save(self.dict_save, save_path + "/best.pth")

        # Save init and epoch or iter
        if is_init:
            torch.save(self.dict_save, save_path + "/init.pth")

        if save_type == "Epoch":
            save_name = "/epoch_iter/%s_%s.pth" % (
                str(self.accumulated_epoch).zfill(5),
                str(0).zfill(5),
            )
            torch.save(self.dict_save, save_path + save_name)
        elif save_type == "Iter":
            save_name = "/epoch_iter/%s_%s.pth" % (
                str(self.accumulated_epoch).zfill(5),
                str(self.curr_iter).zfill(5),
            )
            torch.save(self.dict_save, save_path + save_name)

    def load_dict(self, save_path):
        save_dict = torch.load(save_path)
        default_keys = [
            "accumulated_epoch",
            "accumulated_iter",
            "curr_epoch",
            "curr_iter",
        ]

        for key in default_keys:
            if key not in save_dict.keys():
                raise ValueError("Not supported type of dictionary.")

        for key in [
            "accumulated_epoch",
            "accumulated_iter",
            "curr_epoch",
            "curr_iter",
            "scheduler_type",
        ]:
            setattr(self, key, save_dict[key])
            print("%s is successfully loaded!" % key)

        for key in ["record_info"]:
            self.dict_record = save_dict[key]
            print("%s is successfully loaded!" % key)

        for key in ["recordmanager"]:
            self.rm = save_dict[key]
            print("%s is successfully loaded!" % key)

        for key in ["rmodel", "optimizer", "scheduler", "minimizer"]:
            value = getattr(self, key)
            if value is not None:
                value.load_state_dict(save_dict[key])
                print("%s is successfully loaded!" % key)



# A simple torch style logger
# (C) Wei YANG 2017
# Copied from https://github.com/bearpaw/pytorch-classification/blob/master/utils/logger.py



__all__ = ["Logger", "LoggerMonitor", "savefig"]


# def savefig(fname, dpi=None):
#     dpi = 150 if dpi is None else dpi
#     plt.savefig(fname, dpi=dpi)
#
#
# def plot_overlap(logger, names=None):
#     names = logger.names if names is None else names
#     numbers = logger.numbers
#     for _, name in enumerate(names):
#         x = np.arange(len(numbers[name]))
#         plt.plot(x, np.asarray(numbers[name]))
#     return [logger.title + "(" + name + ")" for name in names]


class Logger:
    """Save training process to log file with simple plot function."""

    def __init__(self, fpath, title=None, resume=False):
        self.file = None
        self.resume = resume
        self.title = "" if title is None else title
        if fpath is not None:
            if resume:
                self.file = open(fpath)
                name = self.file.readline()
                self.names = name.rstrip().split("\t")
                self.numbers = {}
                for _, name in enumerate(self.names):
                    self.numbers[name] = []

                for numbers in self.file:
                    numbers = numbers.rstrip().split("\t")
                    for i in range(0, len(numbers)):
                        self.numbers[self.names[i]].append(float(numbers[i]))
                self.file.close()
                self.file = open(fpath, "a")
            else:
                self.file = open(fpath, "w")

    def set_names(self, names):
        if self.resume:
            pass
        # initialize numbers as empty list
        self.numbers = {}
        self.names = names
        for _, name in enumerate(self.names):
            self.file.write(name)
            self.file.write("\t")
            self.numbers[name] = []
        self.file.write("\n")
        self.file.flush()

    def append(self, numbers):
        assert len(self.names) == len(numbers), "Numbers do not match names"
        for index, num in enumerate(numbers):
            self.file.write(f"{num:.6f}")
            self.file.write("\t")
            self.numbers[self.names[index]].append(num)
        self.file.write("\n")
        self.file.flush()

    #    def plot(self, names=None):
    #        names = self.names if names is None else names
    #        numbers = self.numbers
    #        for _, name in enumerate(names):
    #            x = np.arange(len(numbers[name]))
    #            plt.plot(x, np.asarray(numbers[name]))
    #        plt.legend([self.title + "(" + name + ")" for name in names])
    #        plt.grid(True)

    def close(self):
        if self.file is not None:
            self.file.close()


# class LoggerMonitor:
#     """Load and visualize multiple logs."""
#
#     def __init__(self, paths):
#         """paths is a distionary with {name:filepath} pair"""
#         self.loggers = []
#         for title, path in paths.items():
#             logger = Logger(path, title=title, resume=True)
#             self.loggers.append(logger)
#
#     def plot(self, names=None):
#         plt.figure()
#         plt.subplot(121)
#         legend_text = []
#         for logger in self.loggers:
#             legend_text += plot_overlap(logger, names)
#         plt.legend(
#             legend_text, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0
#         )
#         plt.grid(True)


if __name__ == "__main__":
    # Example
    # logger = Logger('test.txt', resume=True)
    # logger.set_names(['Train loss', 'Valid loss','Test loss'])
    #
    # length = 150
    # t = np.arange(length)
    # train_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1
    # valid_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1
    # test_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1
    #
    # for i in range(0, length):
    #     logger.append([train_loss[i], valid_loss[i], test_loss[i]])
    # logger.plot()
    # plt.show()

    # Example: logger monitor
    paths = {
        "resadvnet20": "test1.txt",
        "resadvnet32": "test2.txt",
        "resadvnet44": "test3.txt",
    }

    field = ["Valid Acc."]

#    monitor = LoggerMonitor(paths)
#    monitor.plot()
#    plt.show()
#    savefig("test.eps")



from ..attack import Attack


class VANILA(Attack):
    r"""
    Vanila version of Attack.
    It just returns the input images.

    Arguments:
        model (nn.Module): model to attack.

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.VANILA(model)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model):
        super().__init__("VANILA", model)
        self.supported_mode = ["default"]

    def forward(self, images, labels=None):
        r"""
        Overridden.
        """

        adv_images = images.clone().detach().to(self.device)

        return adv_images



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
import datetime
import os
from collections import namedtuple

import torch
import torchvision
from torch import nn, optim
from torchvision import transforms

import airtk.defense.fat.fat_attack_generator as attack
from airtk.defense.fat.fat_earlystop import (
    FAT_EARLYSTOP,
)
from airtk.model import ResNet18, WideResNet


class FatTraining:
    def __init__(
        self,
        dataset_name: str,
        model_name: str,
        epochs: int = 120,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        epsilon: float = 0.031,
        num_steps: int = 10,
        step_size: float = 0.007,
        seed: int = 7,
        out_dir: str = "data/",
    ) -> None:
        self._dataset_name = dataset_name

        ArgPrototype = namedtuple(
            "ArgPrototype",
            [
                "net",
                "dataset",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "epsilon",
                "num_steps",
                "step_size",
                "seed",
                "tau",
                "rand_init",
                "omega",
                "dynamictau",
                "depth",
                "width_factor",
                "drop_rate",
                "out_dir",
                "resume",
            ],
        )

        self._args = ArgPrototype(
            model_name,
            dataset_name,
            epochs,
            weight_decay,
            lr,
            momentum,
            epsilon,
            num_steps,
            step_size,
            seed,
            0.0,
            True,
            0.001,
            True,
            34,
            10,
            0.0,
            out_dir,
            "",
        )

    def __call__(self) -> None:
        torch.manual_seed(self._args.seed)

        torch.backends.cudnn.deterministic = False

        torch.backends.cudnn.benchmark = True

        if not os.path.exists(self._args.out_dir):
            os.path.makedirs(self._args.out_dir)

        main_fat(self._dataset_name, self._args)


def train(args, model, train_loader, optimizer, tau):
    starttime = datetime.datetime.now()

    loss_sum = 0

    bp_count = 0

    for data, target in train_loader:
        data, target = data.cuda(), target.cuda()

        # Get friendly adversarial training data via early-stopped PGD

        ES = FAT_EARLYSTOP(model, data)

        output_adv, output_target, output_natural, count = ES.earlystop(
            # model,
            # data,
            target,
            step_size=args.step_size,
            epsilon=args.epsilon,
            perturb_steps=args.num_steps,
            tau=tau,
            randominit_type="uniform_randominit",
            loss_fn="cent",
            rand_init=args.rand_init,
            omega=args.omega,
        )

        bp_count += count

        model.train()

        optimizer.zero_grad()

        output = model(output_adv)

        # calculate standard adversarial training loss

        loss = nn.CrossEntropyLoss(reduction="mean")(output, output_target)

        loss_sum += loss.item()

        loss.backward()

        optimizer.step()

    bp_count_avg = bp_count / len(train_loader.dataset)

    endtime = datetime.datetime.now()

    time = (endtime - starttime).seconds

    return time, loss_sum, bp_count_avg


def adjust_tau(args, epoch, dynamictau):
    tau = args.tau

    if dynamictau:
        if epoch <= 50:
            tau = 0

        elif epoch <= 90:
            tau = 1

        else:
            tau = 2
    return tau


def adjust_learning_rate(args, optimizer, epoch):
    """decrease the learning rate"""
    lr = args.lr

    if epoch >= 60:
        lr = args.lr * 0.1

    if epoch >= 90:
        lr = args.lr * 0.01

    if epoch >= 110:
        lr = args.lr * 0.005

    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def save_checkpoint(args, state, filename="checkpoint.pth.tar"):
    filepath = os.path.join(args.out_dir, filename)

    torch.save(state, filepath)


def main_fat(ds_name, args):
    # setup data loader

    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    # print('==> Load Test Data')

    if ds_name == "cifar10":
        trainset = torchvision.datasets.CIFAR10(
            root="./data", train=True, download=True, transform=transform_train
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=128, shuffle=True, num_workers=2
        )

        testset = torchvision.datasets.CIFAR10(
            root="./data", train=False, download=True, transform=transform_test
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=128, shuffle=False, num_workers=2
        )

    elif ds_name == "cifar100":
        trainset = torchvision.datasets.CIFAR100(
            root="./data", train=True, download=True, transform=transform_train
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=128, shuffle=True, num_workers=2
        )

        testset = torchvision.datasets.CIFAR100(
            root="./data", train=False, download=True, transform=transform_test
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=128, shuffle=False, num_workers=2
        )

    else:
        raise NotImplementedError

    if args.net == "res18":
        model = ResNet18(
            num_classes=100 if args.dataset == "cifar100" else 10,
        ).cuda()

    if args.net == "wideres":
        # e.g., WRN-34-10

        model = WideResNet(
            depth=args.depth,
            num_classes=100 if args.dataset == "cifar100" else 10,
            widen_factor=args.width_factor,
            dropRate=args.drop_rate,
        ).cuda()

    model = torch.nn.DataParallel(model)

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    start_epoch = 0

    # Resume

    if args.resume:
        # resume directly point to checkpoint.pth.tar e.g., --resume='./out-dir/checkpoint.pth.tar'

        # print('==> Friendly Adversarial Training Resuming from checkpoint ..')

        # print(args.resume)

        assert os.path.isfile(args.resume)

        checkpoint = torch.load(args.resume)

        start_epoch = checkpoint["epoch"]

        model.load_state_dict(checkpoint["state_dict"])

        optimizer.load_state_dict(checkpoint["optimizer"])

        # logger_test = Logger(os.path.join(out_dir, 'log_results.txt'), title=title, resume=True)

    # else:

    # print('==> Friendly Adversarial Training')

    # logger_test = Logger(os.path.join(args.out_dir, 'log_results.txt'), title=title)

    # logger_test.set_names(['Epoch', 'Natural Test Acc', 'FGSM Acc', 'PGD20 Acc', 'CW Acc'])

    test_nat_acc = 0

    fgsm_acc = 0

    test_pgd20_acc = 0

    cw_acc = 0

    if ds_name == "cifar10":
        num_classes = 10

    elif ds_name == "cifar100":
        num_classes = 100

    else:
        raise NotImplementedError

    for epoch in range(start_epoch, args.epochs):
        adjust_learning_rate(args, optimizer, epoch + 1)

        train_time, train_loss, bp_count_avg = train(
            args,
            model,
            train_loader,
            optimizer,
            adjust_tau(args, epoch + 1, args.dynamictau),
        )

        ## Evalutions the same as DAT.

        loss, test_nat_acc = attack.eval_clean(model, test_loader)

        loss, fgsm_acc = attack.eval_robust(
            model,
            test_loader,
            num_classes,
            perturb_steps=1,
            epsilon=0.031,
            step_size=0.031,
            loss_fn="cent",
            category="Madry",
            rand_init=True,
        )

        loss, test_pgd20_acc = attack.eval_robust(
            model,
            test_loader,
            num_classes,
            perturb_steps=20,
            epsilon=0.031,
            step_size=0.031 / 4,
            loss_fn="cent",
            category="Madry",
            rand_init=True,
        )

        loss, cw_acc = attack.eval_robust(
            model,
            test_loader,
            num_classes,
            perturb_steps=30,
            epsilon=0.031,
            step_size=0.031 / 4,
            loss_fn="cw",
            category="Madry",
            rand_init=True,
        )

        print(
            "Epoch: [%d | %d] | Train Time: %.2f s | BP Average: %.2f | Natural Test Acc %.2f | FGSM Test Acc %.2f | PGD20 Test Acc %.2f | CW Test Acc %.2f |\n"
            % (
                epoch + 1,
                args.epochs,
                train_time,
                bp_count_avg,
                test_nat_acc,
                fgsm_acc,
                test_pgd20_acc,
                cw_acc,
            )
        )

        # logger_test.append([epoch + 1, test_nat_acc, fgsm_acc, test_pgd20_acc, cw_acc])

        save_checkpoint(
            args,
            {
                "epoch": epoch + 1,
                "state_dict": model.state_dict(),
                "bp_avg": bp_count_avg,
                "test_nat_acc": test_nat_acc,
                "test_pgd20_acc": test_pgd20_acc,
                "optimizer": optimizer.state_dict(),
            },
        )



# None attacks
from .attacks.vanila import VANILA
from .attacks.gn import GN

# Linf attacks
from .attacks.fgsm import FGSM
from .attacks.bim import BIM
from .attacks.rfgsm import RFGSM
from .attacks.pgd import PGD
from .attacks.eotpgd import EOTPGD
from .attacks.ffgsm import FFGSM
from .attacks.tpgd import TPGD
from .attacks.mifgsm import MIFGSM
from .attacks.upgd import UPGD
from .attacks.apgd import APGD
from .attacks.apgdt import APGDT
from .attacks.difgsm import DIFGSM
from .attacks.tifgsm import TIFGSM
from .attacks.jitter import Jitter
from .attacks.nifgsm import NIFGSM
from .attacks.pgdrs import PGDRS
from .attacks.sinifgsm import SINIFGSM
from .attacks.vmifgsm import VMIFGSM
from .attacks.vnifgsm import VNIFGSM
from .attacks.spsa import SPSA
from .attacks.pifgsm import PIFGSM
from .attacks.pifgsmpp import PIFGSMPP

# L2 attacks
from .attacks.cw import CW
from .attacks.pgdl2 import PGDL2
from .attacks.pgdrsl2 import PGDRSL2
from .attacks.deepfool import DeepFool
from .attacks.eaden import EADEN

# L1 attacks
from .attacks.eadl1 import EADL1

# L0 attacks
from .attacks.sparsefool import SparseFool
from .attacks.onepixel import OnePixel
from .attacks.pixle import Pixle
from .attacks.jsma import JSMA

# Linf, L2 attacks
from .attacks.fab import FAB
from .attacks.autoattack import AutoAttack
from .attacks.square import Square

# Wrapper Class
from .wrappers.multiattack import MultiAttack
from .wrappers.lgv import LGV

__version__ = "3.5.1"
__all__ = [
    "VANILA",
    "GN",
    "FGSM",
    "BIM",
    "RFGSM",
    "PGD",
    "EOTPGD",
    "FFGSM",
    "TPGD",
    "MIFGSM",
    "UPGD",
    "APGD",
    "APGDT",
    "DIFGSM",
    "TIFGSM",
    "Jitter",
    "NIFGSM",
    "PGDRS",
    "SINIFGSM",
    "VMIFGSM",
    "VNIFGSM",
    "SPSA",
    "JSMA",
    "EADL1",
    "EADEN",
    "PIFGSM",
    "PIFGSMPP",
    "CW",
    "PGDL2",
    "DeepFool",
    "PGDRSL2",
    "SparseFool",
    "OnePixel",
    "Pixle",
    "FAB",
    "AutoAttack",
    "Square",
    "MultiAttack",
    "LGV",
]
__wrapper__ = [
    "LGV",
    "MultiAttack",
]



import torch
import torch.nn as nn

from ..attack import Attack


class DeepFool(Attack):
    r"""
    'DeepFool: A Simple and Accurate Method to Fool Deep Neural Networks'
    [https://arxiv.org/abs/1511.04599]
    Distance Measure : L2
    Arguments:
        model (nn.Module): model to attack.
        steps (int): number of steps. (Default: 50)
        overshoot (float): parameter for enhancing the noise. (Default: 0.02)
    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.
    Examples::
        >>> attack = torchattacks.DeepFool(model, steps=50, overshoot=0.02)
        >>> adv_images = attack(images, labels)
    """

    def __init__(self, model, steps=50, overshoot=0.02):
        super().__init__("DeepFool", model)
        self.steps = steps
        self.overshoot = overshoot
        self.supported_mode = ["default"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        adv_images, target_labels = self.forward_return_target_labels(
            images, labels
        )
        return adv_images

    def forward_return_target_labels(self, images, labels):
        r"""
        Overridden.
        """
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        batch_size = len(images)
        correct = torch.tensor([True] * batch_size)
        target_labels = labels.clone().detach().to(self.device)
        curr_steps = 0

        adv_images = []
        for idx in range(batch_size):
            image = images[idx : idx + 1].clone().detach()
            adv_images.append(image)

        while (True in correct) and (curr_steps < self.steps):
            for idx in range(batch_size):
                if not correct[idx]:
                    continue
                early_stop, pre, adv_image = self._forward_indiv(
                    adv_images[idx], labels[idx]
                )
                adv_images[idx] = adv_image
                target_labels[idx] = pre
                if early_stop:
                    correct[idx] = False
            curr_steps += 1

        adv_images = torch.cat(adv_images).detach()
        return adv_images, target_labels

    def _forward_indiv(self, image, label):
        image.requires_grad = True
        fs = self.get_logits(image)[0]
        _, pre = torch.max(fs, dim=0)
        if pre != label:
            return (True, pre, image)

        ws = self._construct_jacobian(fs, image)
        image = image.detach()

        f_0 = fs[label]
        w_0 = ws[label]

        wrong_classes = [i for i in range(len(fs)) if i != label]
        f_k = fs[wrong_classes]
        w_k = ws[wrong_classes]

        f_prime = f_k - f_0
        w_prime = w_k - w_0
        value = torch.abs(f_prime) / torch.norm(
            nn.Flatten()(w_prime), p=2, dim=1
        )
        _, hat_L = torch.min(value, 0)

        delta = (
            torch.abs(f_prime[hat_L])
            * w_prime[hat_L]
            / (torch.norm(w_prime[hat_L], p=2) ** 2)
        )

        target_label = hat_L if hat_L < label else hat_L + 1

        adv_image = image + (1 + self.overshoot) * delta
        adv_image = torch.clamp(adv_image, min=0, max=1).detach()
        return (False, target_label, adv_image)

    # https://stackoverflow.com/questions/63096122/pytorch-is-it-possible-to-differentiate-a-matrix
    # torch.autograd.functional.jacobian is only for torch >= 1.5.1
    def _construct_jacobian(self, y, x):
        x_grads = []
        for idx, y_element in enumerate(y):
            if x.grad is not None:
                x.grad.zero_()
            y_element.backward(retain_graph=(False or idx + 1 < len(y)))
            x_grads.append(x.grad.clone().detach())
        return torch.stack(x_grads).reshape(*y.shape, *x.shape)



# (c) 2023 Harry24k
# This code is licensed under the MIT license (see LICENSE.md).
"""
This file has been taken from https://github.com/Harry24k/MAIR/blob/main/mair/attacks/attack.py#L18
"""

import time
from collections import OrderedDict

import torch
from torch.utils.data import DataLoader, TensorDataset


def wrapper_method(func):
    # allows you to add extra functionality to an existing function or method
    def wrapper_func(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        for atk in self.__dict__.get("_attacks").values():
            eval("atk." + func.__name__ + "(*args, **kwargs)")
        return result

    return wrapper_func


class Attack:
    """Base class for MAIR's attacks.

    .. note::
        This class automatically sets the device to the device where the given model is.
        It changes the training mode to eval during the attack process.
        To change this, please see `set_model_training_mode`.
    """

    def __init__(self, name: str, model: torch.nn.Module):
        """Initializes internal attack state.

        :param name: Name of the attack.
        :param model: Model to attack.
        """

        self.attack: str = name
        self._attacks: OrderedDict = OrderedDict()
        self.set_model(model)

        try:
            self.device: torch.device = next(model.parameters()).device
        except Exception:
            self.device: torch.device = None
            print(
                "Failed to set device automatically, please try set_device() manual."
            )

        # Controls attack mode.
        self.attack_mode: str = "default"
        self.supported_mode: list[str] = ["default"]
        self.targeted: bool = False
        self._target_map_function: callable = None

        # Controls when normalization is used.
        self.normalization_used: torch.Tensor = None
        self._normalization_applied: bool = None
        if self.model.__class__.__name__ == "RobModel":
            self._set_rmodel_normalization_used(model)

        # Controls model mode during attack.
        self._model_training: bool = False
        self._batchnorm_training: bool = False
        self._dropout_training: bool = False

    def forward(
        self,
        inputs: torch.Tensor,
        labels: torch.Tensor = None,
        *args,
        **kwargs,
    ):
        """Defines the computation performed at every call.
        Should be overridden by all subclasses.
        """
        raise NotImplementedError

    @wrapper_method
    def set_model(self, model: torch.nn.Module):
        """Sets the model for the attack."""

        self.model: torch.nn.Module = model
        self.model_name: str = model.__class__.__name__

    def get_logits(
        self,
        inputs: torch.Tensor,
        labels: torch.Tensor = None,
        *args,
        **kwargs,
    ):
        """Gets the logits from the model."""

        if self._normalization_applied is False:
            inputs = self.normalize(inputs)
        logits: torch.Tensor = self.model(inputs)
        return logits

    @wrapper_method
    def _set_normalization_applied(self, flag: bool):
        """Sets the normalization applied flag."""

        self._normalization_applied: bool = flag

    @wrapper_method
    def set_device(self, device: torch.device):
        """Sets the device for the attack."""

        self.device: torch.device = device

    @wrapper_method
    def _set_rmodel_normalization_used(self, model: torch.nn.Module):
        """Sets attack normalization for MAIR [https://github.com/Harry24k/MAIR]."""

        mean = getattr(model, "mean", None)
        std = getattr(model, "std", None)
        if (mean is not None) and (std is not None):
            if isinstance(mean, torch.Tensor):
                mean = mean.cpu().numpy()
            if isinstance(std, torch.Tensor):
                std = std.cpu().numpy()
            if (mean != 0).all() or (std != 1).all():
                self.set_normalization_used(mean, std)

    @wrapper_method
    def set_normalization_used(self, mean, std):
        """This function sets the normalization parameters (mean and standard deviation) to be used for normalizing inputs.
        The parameters are stored in a dictionary and reshaped to match the input tensor's shape for broadcasting.


        :param mean: The mean values for each channel.
        :param std: The standard deviation values for each channel.
        """
        self.normalization_used = {}
        n_channels = len(mean)
        mean = torch.tensor(mean).reshape(1, n_channels, 1, 1)
        std = torch.tensor(std).reshape(1, n_channels, 1, 1)
        self.normalization_used["mean"] = mean
        self.normalization_used["std"] = std
        self._set_normalization_applied(True)

    def normalize(self, inputs):
        """This function normalizes the input tensor using the previously set mean and standard deviation.
        The normalization is performed channel-wise.

        :param inputs: The input tensor to be normalized.
        :param torch.Tensor: The normalized input tensor.
        """
        mean = self.normalization_used["mean"].to(inputs.device)
        std = self.normalization_used["std"].to(inputs.device)
        return (inputs - mean) / std

    def inverse_normalize(self, inputs):
        """This function performs the inverse operation of the normalize function.
        It takes a normalized tensor and transforms it back to its original state using the set mean and standard deviation.

        :param inputs: The input tensor to be inverse normalized.

        :return torch.Tensor: The inverse normalized input tensor.
        """
        mean = self.normalization_used["mean"].to(inputs.device)
        std = self.normalization_used["std"].to(inputs.device)
        return inputs * std + mean

    def get_mode(self):
        """Get attack mode."""

        return self.attack_mode

    @wrapper_method
    def set_mode_default(self):
        """Set attack mode as default mode."""
        self.attack_mode = "default"
        self.targeted = False
        print("Attack mode is changed to 'default.'")

    @wrapper_method
    def _set_mode_targeted(self, mode, quiet):
        """This function sets the attack mode to 'targeted' if it is supported.
        If the targeted mode is not supported, it raises a ValueError.
        It also sets the 'targeted' attribute to True.

        :param mode: The attack mode to be set.
        :param quiet: If False, the function will print a message indicating the change in attack mode.

        :raise ValueError: If the targeted mode is not supported.

        """
        if "targeted" not in self.supported_mode:
            raise ValueError("Targeted mode is not supported.")
        self.targeted = True
        self.attack_mode = mode
        if not quiet:
            print("Attack mode is changed to '%s'." % mode)

    @wrapper_method
    def set_mode_targeted_by_function(self, target_map_function, quiet=False):
        """Set attack mode as targeted.

        :param target_map_function: Label mapping function.
                e.g. lambda inputs, labels:(labels+1)%10.
                None for using input labels as targeted labels. (Default)
        :param quiet: Display information message or not. (Default: False)
        """
        self._set_mode_targeted("targeted(custom)", quiet)
        self._target_map_function = target_map_function

    @wrapper_method
    def set_mode_targeted_random(self, quiet=False):
        """Set attack mode as targeted with random labels.

        :param quiet: Display information message or not. (Default: False)
        """

        self._set_mode_targeted("targeted(random)", quiet)
        self._target_map_function = self.get_random_target_label

    @wrapper_method
    def set_mode_targeted_least_likely(self, kth_min=1, quiet=False):
        """Set attack mode as targeted with least likely labels.

        :param kth_min: label with the k-th smallest probability used as target labels. (Default: 1)
        :param num_classses: number of classes. (Default: False)
        """
        self._set_mode_targeted("targeted(least-likely)", quiet)
        assert kth_min > 0
        self._kth_min = kth_min
        self._target_map_function = self.get_least_likely_label

    @wrapper_method
    def set_mode_targeted_by_label(self, quiet=False):
        """Set attack mode as targeted.

        :param quiet: Display information message or not. (Default: False)

        .. note::
            Use user-supplied labels as target labels.
        """
        self._set_mode_targeted("targeted(label)", quiet)
        self._target_map_function = "function is a string"

    @wrapper_method
    def set_model_training_mode(
        self,
        model_training=False,
        batchnorm_training=False,
        dropout_training=False,
    ):
        """Set training mode during attack process.

        :param model_training: True for using training mode for the entire model during attack process.
        :param batchnorm_training: True for using training mode for batchnorms during attack process.
        :param dropout_training: True for using training mode for dropouts during attack process.

        .. note::
            For RNN-based models, we cannot calculate gradients with eval mode.
            Thus, it should be changed to the training mode during the attack.
        """
        self._model_training = model_training
        self._batchnorm_training = batchnorm_training
        self._dropout_training = dropout_training

    @wrapper_method
    def _change_model_mode(self, given_training):
        """This function changes the mode of the model based on the 'given_training' flag.
        If 'given_training' is True, it sets the model to training mode.
        Additionally, it iterates over all modules in the model and sets BatchNorm and Dropout layers to evaluation mode
        if '_batchnorm_training' and '_dropout_training' flags are False, respectively.
        If 'given_training' is False, it sets the model to evaluation mode.

        :PARAM given_training (bool): Flag indicating whether the model should be in training mode or not.
        """
        if self._model_training:
            self.model.train()
            for _, m in self.model.named_modules():
                if not self._batchnorm_training:
                    if "BatchNorm" in m.__class__.__name__:
                        m = m.eval()
                if not self._dropout_training:
                    if "Dropout" in m.__class__.__name__:
                        m = m.eval()
            else:
                self.model.eval()

    @wrapper_method
    def _recover_model_mode(self, given_training):
        """This function recovers the mode of the model based on the 'given_training' flag.
        If 'given_training' is True, it sets the model back to training mode.
        This function is typically used to revert the model back to its original mode after certain operations.

        :param given_training (bool): Flag indicating whether the model should be in training mode or not.
        """
        if given_training:
            self.model.train()

    def save(
        self,
        data_loader,
        save_path=None,
        verbose=True,
        return_verbose=False,
        save_predictions=False,
        save_clean_inputs=False,
        save_type="float",
    ):
        """Save adversarial inputs as torch.tensor from given torch.utils.data.DataLoader.

        :param save_path: save_path.
        :param data_loader: data loader.
        :param verbose: True for displaying detailed information. (Default: True)
        :param return_verbose: True for returning detailed information. (Default: False)
        :param save_predictions: True for saving predicted labels (Default: False)
        :param save_clean_inputs: True for saving clean inputs (Default: False)
        """
        if save_path is not None:
            adv_input_list = []
            label_list = []
            if save_predictions:
                pred_list = []
            if save_clean_inputs:
                input_list = []

        correct = 0
        total = 0
        l2_distance = []

        total_batch = len(data_loader)
        given_training = self.model.training

        for step, (inputs, labels) in enumerate(data_loader):
            start = time.time()
            adv_inputs = self.__call__(inputs, labels)
            batch_size = len(inputs)

            if verbose or return_verbose:
                with torch.no_grad():
                    outputs = self.get_output_with_eval_nograd(adv_inputs)

                    # Calculate robust accuracy
                    _, pred = torch.max(outputs.data, 1)
                    total += labels.size(0)
                    right_idx = pred == labels.to(self.device)
                    correct += right_idx.sum()
                    rob_acc = 100 * float(correct) / total

                    # Calculate l2 distance
                    delta = (adv_inputs - inputs.to(self.device)).view(
                        batch_size, -1
                    )  # nopep8
                    l2_distance.append(
                        torch.norm(delta[~right_idx], p=2, dim=1)
                    )  # nopep8
                    l2 = torch.cat(l2_distance).mean().item()

                    # Calculate time computation
                    progress = (step + 1) / total_batch * 100
                    end = time.time()
                    elapsed_time = end - start

                    if verbose:
                        self._save_print(
                            progress, rob_acc, l2, elapsed_time, end="\r"
                        )  # nopep8

            if save_path is not None:
                adv_input_list.append(adv_inputs.detach().cpu())
                label_list.append(labels.detach().cpu())

                adv_input_list_cat = torch.cat(adv_input_list, 0)
                label_list_cat = torch.cat(label_list, 0)

                save_dict = {
                    "adv_inputs": adv_input_list_cat,
                    "labels": label_list_cat,
                }  # nopep8

                if save_predictions:
                    pred_list.append(pred.detach().cpu())
                    pred_list_cat = torch.cat(pred_list, 0)
                    save_dict["preds"] = pred_list_cat

                if save_clean_inputs:
                    input_list.append(inputs.detach().cpu())
                    input_list_cat = torch.cat(input_list, 0)
                    save_dict["clean_inputs"] = input_list_cat

                if self.normalization_used is not None:
                    save_dict["adv_inputs"] = self.inverse_normalize(
                        save_dict["adv_inputs"]
                    )  # nopep8
                    if save_clean_inputs:
                        save_dict["clean_inputs"] = self.inverse_normalize(
                            save_dict["clean_inputs"]
                        )  # nopep8

                if save_type == "int":
                    save_dict["adv_inputs"] = self.to_type(
                        save_dict["adv_inputs"], "int"
                    )  # nopep8
                    if save_clean_inputs:
                        save_dict["clean_inputs"] = self.to_type(
                            save_dict["clean_inputs"], "int"
                        )  # nopep8

                save_dict["save_type"] = save_type
                torch.save(save_dict, save_path)

        # To avoid erasing the printed information.
        if verbose:
            self._save_print(progress, rob_acc, l2, elapsed_time, end="\n")

        if given_training:
            self.model.train()

        if return_verbose:
            return rob_acc, l2, elapsed_time

    @staticmethod
    def to_type(inputs, type):
        """Return inputs as int if float is given."""
        if type == "int":
            if isinstance(inputs, torch.FloatTensor) or isinstance(
                inputs, torch.cuda.FloatTensor
            ):
                return (inputs * 255).type(torch.uint8)
        elif type == "float":
            if isinstance(inputs, torch.ByteTensor) or isinstance(
                inputs, torch.cuda.ByteTensor
            ):
                return inputs.float() / 255
        else:
            raise ValueError(
                type + " is not a valid type. [Options: float, int]"
            )
        return inputs

    @staticmethod
    def _save_print(progress, rob_acc, l2, elapsed_time, end):
        """This function prints the progress of a certain operation, robust accuracy, L2 norm, and elapsed time per iteration.
        It's typically used for logging purposes during training or testing of models.

        :param progress: The progress of the operation in percentage.
        :param rob_acc: The robust accuracy of the model in percentage.
        :param l2: The L2 norm of the model parameters or gradients.
        :param elapsed_time: The time elapsed per iteration in seconds.
        :param end: The end character to be used in the print function. Typically '\n' for a new line or '\r' to overwrite the current line.
        """
        print(
            "- Save progress: %2.2f %% / Robust accuracy: %2.2f %% / L2: %1.5f (%2.3f it/s) \t"
            % (progress, rob_acc, l2, elapsed_time),
            end=end,
        )

    @staticmethod
    def load(
        load_path,
        batch_size=128,
        shuffle=False,
        normalize=None,
        load_predictions=False,
        load_clean_inputs=False,
    ):
        save_dict = torch.load(load_path)
        keys = ["adv_inputs", "labels"]

        if load_predictions:
            keys.append("preds")
        if load_clean_inputs:
            keys.append("clean_inputs")

        if save_dict["save_type"] == "int":
            save_dict["adv_inputs"] = save_dict["adv_inputs"].float() / 255
            if load_clean_inputs:
                save_dict["clean_inputs"] = (
                    save_dict["clean_inputs"].float() / 255
                )  # nopep8

        if normalize is not None:
            n_channels = len(normalize["mean"])
            mean = torch.tensor(normalize["mean"]).reshape(1, n_channels, 1, 1)
            std = torch.tensor(normalize["std"]).reshape(1, n_channels, 1, 1)
            save_dict["adv_inputs"] = (save_dict["adv_inputs"] - mean) / std
            if load_clean_inputs:
                save_dict["clean_inputs"] = (
                    save_dict["clean_inputs"] - mean
                ) / std  # nopep8

        adv_data = TensorDataset(*[save_dict[key] for key in keys])
        adv_loader = DataLoader(
            adv_data, batch_size=batch_size, shuffle=shuffle
        )
        print(
            "Data is loaded in the following order: [%s]" % (", ".join(keys))
        )  # nopep8
        return adv_loader

    @torch.no_grad()
    def get_output_with_eval_nograd(self, inputs):
        """This function loads a saved dataset from a given path and returns a DataLoader object.
        The function also supports optional normalization and loading of additional data such as predictions and clean inputs.

        :param load_path: The path to the saved dataset.
        :param batch_size: The batch size for the DataLoader. Defaults to 128.
        :param shuffle: Whether to shuffle the data. Defaults to False.
        :param normalize: A dictionary containing 'mean' and 'std' for normalization. Defaults to None.
        :param load_predictions: Whether to load predictions. Defaults to False.
        :param load_clean_inputs: Whether to load clean inputs. Defaults to False.

        :return DataLoader: A DataLoader object containing the loaded data.
        """
        given_training = self.model.training
        if given_training:
            self.model.eval()
        outputs = self.get_logits(inputs)
        if given_training:
            self.model.train()
        return outputs

    def get_target_label(self, inputs, labels=None):
        """Function for changing the attack mode.

        :return  input labels.
        """
        if self._target_map_function is None:
            raise ValueError(
                "target_map_function is not initialized by set_mode_targeted."
            )
        if self.attack_mode == "targeted(label)":
            target_labels = labels
        else:
            target_labels = self._target_map_function(inputs, labels)
        return target_labels

    @torch.no_grad()
    def get_least_likely_label(self, inputs, labels=None):
        """This function computes the least likely label for each input sample based on the model's output.
        If labels are not provided, it uses the model's predictions as labels.
        It's typically used in adversarial attack scenarios where the least likely label is targeted.

        :param inputs: The input samples.
        :param labels: The true labels of the input samples. Defaults to None.

        :return torch.Tensor: The least likely labels for each input sample.
        """
        outputs = self.get_output_with_eval_nograd(inputs)
        if labels is None:
            _, labels = torch.max(outputs, dim=1)
        n_classses = outputs.shape[-1]

        target_labels = torch.zeros_like(labels)
        for counter in range(labels.shape[0]):
            l = list(range(n_classses))
            l.remove(labels[counter])
            _, t = torch.kthvalue(outputs[counter][l], self._kth_min)
            target_labels[counter] = l[t]

        return target_labels.long().to(self.device)

    @torch.no_grad()
    def get_random_target_label(self, inputs, labels=None):
        """
        This function computes a random target label for each input sample that is different from its true label.
        If labels are not provided, it uses the model's predictions as labels.
        It's typically used in adversarial attack scenarios where a random label is targeted.

        :param inputs: The input samples.
        :param labels: The true labels of the input samples. Defaults to None.

        :return torch.Tensor: The random target labels for each input sample.
        """
        outputs = self.get_output_with_eval_nograd(inputs)
        if labels is None:
            _, labels = torch.max(outputs, dim=1)
        n_classses = outputs.shape[-1]

        target_labels = torch.zeros_like(labels)
        for counter in range(labels.shape[0]):
            l = list(range(n_classses))
            l.remove(labels[counter])
            t = (len(l) * torch.rand([1])).long().to(self.device)
            target_labels[counter] = l[t]

        return target_labels.long().to(self.device)

    def __call__(self, inputs, labels=None, *args, **kwargs):
        """This function is the main entry point for this class. It's called when an instance of the class is invoked like a function.
        It performs the following steps:
        1. Saves the current training mode of the model.
        2. Changes the model's mode based on the saved training mode.
        3. If normalization has been applied, it inverses the normalization on the inputs and sets the '_normalization_applied' flag to False.
        4. Calls the 'forward' method to perform the main operation of the class (e.g., generating adversarial examples).
        5. If normalization has been applied, it normalizes the adversarial inputs and sets the '_normalization_applied' flag back to True.
        6. Recovers the model's mode to its original state.
        7. Returns the adversarial inputs.

        :param inputs: The input samples.
        :param labels: The true labels of the input samples. Defaults to None.
        :param *args: Variable length argument list.
        :param **kwargs: Arbitrary keyword arguments.

        :return torch.Tensor: The adversarial inputs.
        """
        given_training = self.model.training
        self._change_model_mode(given_training)

        if self._normalization_applied is True:
            inputs = self.inverse_normalize(inputs)
            self._set_normalization_applied(False)

            adv_inputs = self.forward(inputs, labels, *args, **kwargs)
            # adv_inputs = self.to_type(adv_inputs, self.return_type)

            adv_inputs = self.normalize(adv_inputs)
            self._set_normalization_applied(True)
        else:
            adv_inputs = self.forward(inputs, labels, *args, **kwargs)
            # adv_inputs = self.to_type(adv_inputs, self.return_type)

        self._recover_model_mode(given_training)

        return adv_inputs

    def __repr__(self):
        """This function provides a string representation of the object.
        It's typically used for debugging and logging purposes.
        The function copies the object's attributes dictionary, removes certain keys,
        and then formats the remaining items into a string.

        :return A string representation of the object.
        """
        info = self.__dict__.copy()

        del_keys = ["model", "attack", "supported_mode"]

        for key in info.keys():
            if key[0] == "_":
                del_keys.append(key)

        for key in del_keys:
            del info[key]

        info["attack_mode"] = self.attack_mode
        info["normalization_used"] = (
            True if self.normalization_used is not None else False
        )

        return (
            self.attack
            + "("
            + ", ".join(f"{key}={val}" for key, val in info.items())
            + ")"
        )

    def __setattr__(self, name, value):
        """This function overrides the default behavior of the 'setattr' function.
        It's called when an attribute value is set.
        Besides setting the attribute value, it also updates the '_attacks' dictionary
        if the value is an instance of the 'Attack' class or contains instances of the 'Attack' class.
        :param name: The name of the attribute.
        :param value: The value of the attribute.
        """
        object.__setattr__(self, name, value)

        attacks = self.__dict__.get("_attacks")

        # Define a helper function to get all values in iterable items
        def get_all_values(items, stack=[]):
            if items not in stack:
                stack.append(items)
                if isinstance(items, list) or isinstance(items, dict):
                    if isinstance(items, dict):
                        items = list(items.keys()) + list(items.values())
                    for item in items:
                        yield from get_all_values(item, stack)
                else:
                    if isinstance(items, Attack):
                        yield items
            else:
                if isinstance(items, Attack):
                    yield items

        # For each 'Attack' instance in the value, add it to the '_attacks' dictionary
        for num, value in enumerate(get_all_values(value)):
            attacks[name + "." + str(num)] = value
            for subname, subvalue in value.__dict__.get("_attacks").items():
                attacks[name + "." + subname] = subvalue



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import numpy as np
import torch
from torch import nn
from torch.autograd import Variable


class FAT_ATTACK_GENERATOR:
    def __init__(self, model, device):
        self.model = model.to(device)
        self.device = device

    def cwloss(self, output, target, num_classes, confidence=50):
        """
        This function computes the Carlini-Wagner loss between the output of a model and the target labels.
        The CW loss is used for generating adversarial examples.
        The implementation as in repo CAT https://github.com/sunblaze-ucb/curriculum-adversarial-training-CAT

        :param output: The output tensor from the model. It should have shape (batch_size, num_classes).
        :param target: The true labels for the data. It should be a 1D tensor of shape (batch_size,).
        :param num_classes: The number of classes in the classification task.
        :param confidence: The confidence margin, which is added to increase the confidence of the adversarial example
        :param device: to set all the tensors to the same device.

        :return: The computed CW loss.
        """
        device = self.device
        target = target.to(device)
        output = output.to(device)
        # Convert the target to its one-hot representation
        target = target.data
        target_onehot = torch.zeros(target.size() + (num_classes,))
        target_onehot = target_onehot.to(device)
        target_onehot.scatter_(1, target.unsqueeze(1), 1.0)

        # Create a variable for the target
        target_var = Variable(target_onehot, requires_grad=False)

        # Compute the real class probabilities
        real = (target_var * output).sum(1)

        # Compute the maximum other class probabilities
        other = ((1.0 - target_var) * output - target_var * 10000.0).max(1)[0]

        # Compute the CW loss
        loss = -torch.clamp(
            real - other + confidence, min=0.0
        )  # equiv to max(..., 0.)
        loss = torch.sum(loss)

        return loss

    def pgd(
        self,
        data,
        target,
        epsilon,
        step_size,
        num_steps,
        num_classes,
        loss_fn,
        category,
        rand_init,
    ):
        """
        This function implements the Projected Gradient Descent (PGD) attack.

        :param model: The model to attack.
        :param data: The input data.
        :param target: The target labels.
        :param epsilon: The maximum perturbation for each pixel.
        :param step_size: The step size for each update.
        :param num_steps: The number of steps for the attack.
        :param num_classes: The number of classes in the classification task.
        :param loss_fn: The loss function to use ('cent' for CrossEntropyLoss, 'cw' for CW loss).
        :param category: The category of the attack ('trades' or 'Madry').
        :param rand_init: Whether to initialize the perturbation with random noise.
        :param device: The device to perform computations on.

        :return: The adversarial examples.
        """
        # Move the model and data to the specified device
        device = self.device
        model = self.model
        data = data.to(device)
        target = target.to(device)

        # Set the model to evaluation mode
        model.eval()

        # Initialize the adversarial examples
        if category == "trades":
            x_adv = (
                data.detach()
                + 0.001 * torch.randn(data.shape).to(device).detach()
                if rand_init
                else data.detach()
            )
        elif category == "Madry":
            x_adv = (
                data.detach()
                + torch.from_numpy(
                    np.random.uniform(-epsilon, epsilon, data.shape)
                )
                .float()
                .to(device)
                if rand_init
                else data.detach()
            )
            x_adv = torch.clamp(x_adv, 0.0, 1.0)

        # Perform the attack
        for _ in range(num_steps):
            x_adv.requires_grad_()
            output = model(x_adv)
            model.zero_grad()
            with torch.enable_grad():
                if loss_fn == "cent":
                    loss_adv = nn.CrossEntropyLoss(reduction="mean")(
                        output, target
                    )
                if loss_fn == "cw":
                    loss_adv = self.cwloss(output, target, num_classes)
            loss_adv.backward()
            eta = (
                step_size * x_adv.grad.sign()
            )  # (Pylance) not being able to infer the type of x_adv
            x_adv = x_adv.detach() + eta
            x_adv = torch.min(torch.max(x_adv, data - epsilon), data + epsilon)
            x_adv = torch.clamp(x_adv, 0.0, 1.0)

        return x_adv

    def eval_clean(self, test_loader):
        """
        This function evaluates a model on clean (non-adversarial) data.

        :param model: The model to evaluate.
        :param test_loader: The DataLoader for the test data.
        :param device: The device to perform computations on.

        :return: The average test loss and test accuracy.
        """
        device = self.device
        # Move the model to the specified device
        model = self.model
        model = model.to(device)

        # Set the model to evaluation mode
        model.eval()

        # Initialize the test loss and the number of correct predictions
        test_loss = 0
        correct = 0

        # Disable gradient computations
        with torch.no_grad():
            # Iterate over the test data
            for data, target in test_loader:
                # Move the data and target to the specified device
                data, target = data.to(device), target.to(device)

                # Get the output of the model
                output = model(data)

                # Compute the CrossEntropyLoss and add it to the total test loss
                test_loss += nn.CrossEntropyLoss(reduction="mean")(
                    output, target
                ).item()

                # Get the predicted class for each sample
                pred = output.max(1, keepdim=True)[1]

                # Count the number of correct predictions
                correct += pred.eq(target.view_as(pred)).sum().item()

        # Compute the average test loss
        test_loss /= len(test_loader.dataset)

        # Compute the test accuracy
        test_accuracy = correct / len(test_loader.dataset)

        return test_loss, test_accuracy

    def eval_robust(
        self,
        test_loader,
        num_classes,
        perturb_steps,
        epsilon,
        step_size,
        loss_fn,
        category,
        rand_init,
    ):
        """
        This function evaluates the robustness of a model against adversarial attacks.

        :param model: The model to evaluate.
        :param test_loader: The DataLoader for the test data.
        :param num_classes: The number of classes in the classification task.
        :param perturb_steps: The number of steps for the attack.
        :param epsilon: The maximum perturbation for each pixel.
        :param step_size: The step size for each update.
        :param loss_fn: The loss function to use ('cent' for CrossEntropyLoss, 'cw' for CW loss).
        :param category: The category of the attack ('trades' or 'Madry').
        :param rand_init: Whether to initialize the perturbation with random noise.
        :param device: The device to perform computations on.
        :return: The average test loss and test accuracy.
        """
        device = self.device
        # Move the model to the specified device
        model = self.model

        # Set the model to evaluation mode
        model.eval()

        # Initialize the test loss and the number of correct predictions
        test_loss = 0
        correct = 0

        # Enable gradient computations
        with torch.enable_grad():
            # Iterate over the test data
            for data, target in test_loader:
                # Move the data and target to the specified device
                data, target = data.to(device), target.to(device)

                # Generate adversarial examples using the PGD attack
                x_adv = self.pgd(
                    data,
                    target,
                    epsilon,
                    step_size,
                    perturb_steps,
                    num_classes,
                    loss_fn,
                    category,
                    rand_init=rand_init,
                )

                # Get the output of the model on the adversarial examples
                output = model(x_adv)

                # Compute the CrossEntropyLoss and add it to the total test loss
                test_loss += nn.CrossEntropyLoss(reduction="mean")(
                    output, target
                ).item()

                # Get the predicted class for each sample
                pred = output.max(1, keepdim=True)[1]

                # Count the number of correct predictions
                correct += pred.eq(target.view_as(pred)).sum().item()

        # Compute the average test loss
        test_loss /= len(test_loader.dataset)

        # Compute the test accuracy
        test_accuracy = correct / len(test_loader.dataset)

        return test_loss, test_accuracy
